# Copyright 2014-2020 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Oliver J. Backhouse <olbackhouse@gmail.com>
#         George H. Booth <george.booth@kcl.ac.uk>
#

'''
Auxiliary second-order Green's function perturbation theory for
unrestricted references
'''

import time
import numpy as np
from pyscf import lib
from pyscf.lib import logger
from pyscf import __config__
from pyscf import ao2mo
from pyscf.scf import _vhf
from pyscf.agf2 import ragf2, _agf2, mpi_helper
from pyscf.agf2 import aux_space as aux
from pyscf.agf2.chempot import binsearch_chempot, minimize_chempot
from pyscf.mp.ump2 import get_frozen_mask as _get_frozen_mask

BLKMIN = getattr(__config__, 'agf2_blkmin', 1)


def build_se_part(agf2, eri, gf_occ, gf_vir, os_factor=1.0, ss_factor=1.0):
    ''' Builds either the auxiliaries of the occupied self-energy,
        or virtual if :attr:`gf_occ` and :attr:`gf_vir` are swapped,
        for a single spin.

    Args:
        eri : _ChemistsERIs
            Electronic repulsion integrals
        gf_occ : tuple of GreensFunction
            Occupied Green's function for each spin
        gf_vir : tuple of GreensFunction
            Virtual Green's function for each spin

    Kwargs:
        os_factor : float
            Opposite-spin factor for spin-component-scaled (SCS)
            calculations. Default 1.0
        ss_factor : float
            Same-spin factor for spin-component-scaled (SCS)
            calculations. Default 1.0

    Returns:
        :class:`SelfEnergy`
    '''

    cput0 = (logger.process_clock(), logger.perf_counter())
    log = logger.Logger(agf2.stdout, agf2.verbose)

    assert type(gf_occ[0]) is aux.GreensFunction
    assert type(gf_occ[1]) is aux.GreensFunction
    assert type(gf_vir[0]) is aux.GreensFunction
    assert type(gf_vir[1]) is aux.GreensFunction

    nmo = eri.nmo
    noa, nob = gf_occ[0].naux, gf_occ[1].naux
    nva, nvb = gf_vir[0].naux, gf_vir[1].naux
    tol = agf2.weight_tol
    facs = {'os_factor': os_factor, 'ss_factor': ss_factor}

    ci_a, ei_a = gf_occ[0].coupling, gf_occ[0].energy
    ci_b, ei_b = gf_occ[1].coupling, gf_occ[1].energy
    ca_a, ea_a = gf_vir[0].coupling, gf_vir[0].energy
    ca_b, ea_b = gf_vir[1].coupling, gf_vir[1].energy

    mem_incore = (nmo[0]*noa*(noa*nva+nob*nvb)) * 8/1e6
    mem_now = lib.current_memory()[0]
    if (mem_incore+mem_now < agf2.max_memory) or agf2.incore_complete:
        qeri = _make_qmo_eris_incore(agf2, eri, (ci_a, ci_a, ca_a), (ci_b, ci_b, ca_b), spin=0)
    else:
        qeri = _make_qmo_eris_outcore(agf2, eri, (ci_a, ci_a, ca_a), (ci_b, ci_b, ca_b), spin=0)

    if isinstance(qeri[0], np.ndarray):
        vv, vev = _agf2.build_mats_uagf2_incore(qeri, (ei_a, ei_b), (ea_a, ea_b), **facs)
    else:
        vv, vev = _agf2.build_mats_uagf2_outcore(qeri, (ei_a, ei_b), (ea_a, ea_b), **facs)

    e, c = _agf2.cholesky_build(vv, vev)
    se_a = aux.SelfEnergy(e, c, chempot=gf_occ[0].chempot)
    se_a.remove_uncoupled(tol=tol)

    if not (agf2.frozen is None or agf2.frozen == 0):
        mask = get_frozen_mask(agf2)
        coupling = np.zeros((nmo[0], se_a.naux))
        coupling[mask[0]] = se_a.coupling
        se_a = aux.SelfEnergy(se_a.energy, coupling, chempot=se_a.chempot)

    cput0 = log.timer('se part (alpha)', *cput0)

    mem_incore = (nmo[1]*nob*(nob*nvb+noa*nva)) * 8/1e6
    mem_now = lib.current_memory()[0]
    if (mem_incore+mem_now < agf2.max_memory) or agf2.incore_complete:
        qeri = _make_qmo_eris_incore(agf2, eri, (ci_a, ci_a, ca_a), (ci_b, ci_b, ca_b), spin=1)
    else:
        qeri = _make_qmo_eris_outcore(agf2, eri, (ci_a, ci_a, ca_a), (ci_b, ci_b, ca_b), spin=1)

    if isinstance(qeri[0], np.ndarray):
        vv, vev = _agf2.build_mats_uagf2_incore(qeri, (ei_b, ei_a), (ea_b, ea_a), **facs)
    else:
        vv, vev = _agf2.build_mats_uagf2_outcore(qeri, (ei_b, ei_a), (ea_b, ea_a), **facs)

    e, c = _agf2.cholesky_build(vv, vev)
    se_b = aux.SelfEnergy(e, c, chempot=gf_occ[1].chempot)
    se_b.remove_uncoupled(tol=tol)

    if not (agf2.frozen is None or agf2.frozen == 0):
        mask = get_frozen_mask(agf2)
        coupling = np.zeros((nmo[1], se_b.naux))
        coupling[mask[1]] = se_b.coupling
        se_b = aux.SelfEnergy(se_b.energy, coupling, chempot=se_b.chempot)

    cput0 = log.timer('se part (beta)', *cput0)

    return (se_a, se_b)


def get_fock(agf2, eri, gf=None, rdm1=None):
    ''' Computes the physical space Fock matrix in MO basis. If :attr:`rdm1`
        is not supplied, it is built from :attr:`gf`, which defaults to
        the mean-field Green's function.

    Args:
        eri : _ChemistsERIs
            Electronic repulsion integrals

    Kwargs:
        gf : GreensFunction
            Auxiliaries of the Green's function
        rdm1 : 2D array
            Reduced density matrix

    Returns:
        ndarray of physical space Fock matrix
    '''

    if rdm1 is None:
        rdm1 = agf2.make_rdm1(gf)

    vj_aa, vk_aa = agf2.get_jk(eri.eri_aa, rdm1=rdm1[0])
    vj_bb, vk_bb = agf2.get_jk(eri.eri_bb, rdm1=rdm1[1])
    vj_ab = agf2.get_jk(eri.eri_ab, rdm1=rdm1[1], with_k=False)[0]
    vj_ba = agf2.get_jk(eri.eri_ba, rdm1=rdm1[0], with_k=False)[0]

    fock_a = eri.h1e[0] + vj_aa + vj_ab - vk_aa
    fock_b = eri.h1e[1] + vj_bb + vj_ba - vk_bb

    fock = (fock_a, fock_b)

    return fock


def fock_loop(agf2, eri, gf, se):
    ''' Self-consistent loop for the density matrix via the HF self-
        consistent field.

    Args:
        eri : _ChemistsERIs
            Electronic repulsion integrals
        gf : tuple of GreensFunction
            Auxiliaries of the Green's function for each spin
        se : tuple of SelfEnergy
            Auxiliaries of the self-energy for each spin

    Returns:
        :class:`SelfEnergy`, :class:`GreensFunction` and a boolean
        indicating whether convergence was successful.
    '''

    assert type(gf[0]) is aux.GreensFunction
    assert type(gf[1]) is aux.GreensFunction
    assert type(se[0]) is aux.SelfEnergy
    assert type(se[1]) is aux.SelfEnergy

    cput0 = cput1 = (logger.process_clock(), logger.perf_counter())
    log = logger.Logger(agf2.stdout, agf2.verbose)

    diis = lib.diis.DIIS(agf2)
    diis.space = agf2.fock_diis_space
    diis.min_space = agf2.fock_diis_min_space
    focka, fockb = agf2.get_fock(eri, gf)
    sea, seb = se
    gfa, gfb = gf

    nalph, nbeta = agf2.nocc
    nmoa, nmob = eri.nmo
    nauxa, nauxb = sea.naux, seb.naux
    nqmoa, nqmob = nauxa+nmoa, nauxb+nmob
    bufa, bufb = np.zeros((nqmoa, nqmoa)), np.zeros((nqmob, nqmob))
    rdm1a_prev = 0
    rdm1b_prev = 0
    converged = False
    opts = {'tol': agf2.conv_tol_nelec, 'maxiter': agf2.max_cycle_inner}

    for niter1 in range(1, agf2.max_cycle_outer+1):
        sea, opt = minimize_chempot(sea, focka, nalph, x0=sea.chempot,
                                    occupancy=1, **opts)
        seb, opt = minimize_chempot(seb, fockb, nbeta, x0=seb.chempot,
                                    occupancy=1, **opts)

        for niter2 in range(1, agf2.max_cycle_inner+1):
            wa, va = sea.eig(focka, chempot=0.0, out=bufa)
            wb, vb = seb.eig(fockb, chempot=0.0, out=bufb)
            sea.chempot, nerra = \
                    binsearch_chempot((wa, va), nmoa, nalph, occupancy=1)
            seb.chempot, nerrb = \
                    binsearch_chempot((wb, vb), nmob, nbeta, occupancy=1)
            nerr = max(nerra, nerrb)

            wa, va = sea.eig(focka, out=bufa)
            wb, vb = seb.eig(fockb, out=bufb)
            gfa = aux.GreensFunction(wa, va[:nmoa], chempot=sea.chempot)
            gfb = aux.GreensFunction(wb, vb[:nmob], chempot=seb.chempot)

            gf = (gfa, gfb)
            focka, fockb = agf2.get_fock(eri, gf)
            rdm1a, rdm1b = agf2.make_rdm1(gf)
            focka, fockb = diis.update(np.array((focka, fockb)), xerr=None)

            if niter2 > 1:
                derra = np.max(np.absolute(rdm1a - rdm1a_prev))
                derrb = np.max(np.absolute(rdm1b - rdm1b_prev))
                derr = max(derra, derrb)

                if derr < agf2.conv_tol_rdm1:
                    break

            rdm1a_prev = rdm1a.copy()
            rdm1b_prev = rdm1b.copy()

        log.debug1('fock loop %d  cycles = %d  dN = %.3g  |ddm| = %.3g',
                   niter1, niter2, nerr, derr)
        cput1 = log.timer_debug1('fock loop %d'%niter1, *cput1)

        if derr < agf2.conv_tol_rdm1 and abs(nerr) < agf2.conv_tol_nelec:
            converged = True
            break

    se = (sea, seb)

    log.info('fock converged = %s' % converged)
    log.info(' alpha: chempot = %.9g  dN = %.3g  |ddm| = %.3g',
             sea.chempot, nerra, derra)
    log.info(' beta:  chempot = %.9g  dN = %.3g  |ddm| = %.3g',
             seb.chempot, nerrb, derrb)
    log.timer('fock loop', *cput0)

    return gf, se, converged


def energy_1body(agf2, eri, gf):
    ''' Calculates the one-body energy according to the UHF form.

    Args:
        eri : _ChemistsERIs
            Electronic repulsion integrals
        gf : tuple of GreensFunction
            Auxiliaries of the Green's function for each spin

    Returns:
        One-body energy
    '''

    assert type(gf[0]) is aux.GreensFunction
    assert type(gf[1]) is aux.GreensFunction

    rdm1 = agf2.make_rdm1(gf)
    fock = agf2.get_fock(eri, gf)

    e1b_a = 0.5 * np.sum(rdm1[0] * (eri.h1e[0] + fock[0]))
    e1b_b = 0.5 * np.sum(rdm1[1] * (eri.h1e[1] + fock[1]))

    e1b = e1b_a + e1b_b
    e1b += agf2.energy_nuc()

    return e1b


def energy_2body(agf2, gf, se):
    ''' Calculates the two-body energy using analytically integrated
        Galitskii-Migdal formula. The formula is symmetric and only
        one side needs to be calculated.

    Args:
        gf : tuple of GreensFunction
            Auxiliaries of the Green's function for each spin
        se : tuple of SelfEnergy
            Auxiliaries of the self-energy for each spin

    Returns:
        Two-body energy
    '''

    e2b_a = ragf2.energy_2body(agf2, gf[0], se[0])
    e2b_b = ragf2.energy_2body(agf2, gf[1], se[1])

    e2b = (e2b_a + e2b_b) * 0.5

    return e2b


def energy_mp2(agf2, gf, se):
    ''' Calculates the two-body energy using analytically integrated
        Galitskii-Migdal formula for an MP2 self-energy. Per the
        definition of one- and two-body partitioning in the Dyson
        equation, this result is half of :func:`energy_2body`.

    Args:
        gf : tuple of GreensFunction
            Auxiliaries of the Green's function for each spin
        se : tuple of SelfEnergy
            Auxiliaries of the self-energy for each spin

    Returns:
        MP2 energy
    '''

    emp2_a = ragf2.energy_mp2(agf2, gf[0], se[0])
    emp2_b = ragf2.energy_mp2(agf2, gf[1], se[1])

    emp2 = (emp2_a + emp2_b) * 0.5

    return emp2


class UAGF2(ragf2.RAGF2):
    ''' Unrestricted AGF2 with canonical HF reference

    Attributes:
        verbose : int
            Print level. Default value equals to :class:`Mole.verbose`
        max_memory : float or int
            Allowed memory in MB. Default value equals to :class:`Mole.max_memory`
        incore_complete : bool
            Avoid all I/O. Default is False.
        conv_tol : float
            Convergence threshold for AGF2 energy. Default value is 1e-7
        conv_tol_rdm1 : float
            Convergence threshold for first-order reduced density matrix.
            Default value is 1e-8.
        conv_tol_nelec : float
            Convergence threshold for the number of electrons. Default
            value is 1e-6.
        max_cycle : int
            Maximum number of AGF2 iterations. Default value is 50.
        max_cycle_outer : int
            Maximum number of outer Fock loop iterations. Default
            value is 20.
        max_cycle_inner : int
            Maximum number of inner Fock loop iterations. Default
            value is 50.
        weight_tol : float
            Threshold in spectral weight of auxiliaries to be considered
            zero. Default 1e-11.
        diis : bool or lib.diis.DIIS
            Whether to use DIIS, can also be a lib.diis.DIIS object. Default
            value is True.
        diis_space : int
            DIIS space size. Default value is 8.
        diis_min_space : int
            Minimum space of DIIS. Default value is 1.
        fock_diis_space : int
            DIIS space size for Fock loop iterations. Default value is 6.
        fock_diis_min_space :
            Minimum space of DIIS. Default value is 1.
        os_factor : float
            Opposite-spin factor for spin-component-scaled (SCS)
            calculations. Default 1.0
        ss_factor : float
            Same-spin factor for spin-component-scaled (SCS)
            calculations. Default 1.0
        damping : float
            Damping factor for the self-energy. Default value is 0.0

    Saved results

        e_corr : float
            AGF2 correlation energy
        e_tot : float
            Total energy (HF + correlation)
        e_1b : float
            One-body part of :attr:`e_tot`
        e_2b : float
            Two-body part of :attr:`e_tot`
        e_init : float
            Initial correlation energy (truncated MP2)
        converged : bool
            Whether convergence was successful
        se : tuple of SelfEnergy
            Auxiliaries of the self-energy for each spin
        gf : tuple of GreensFunction
            Auxiliaries of the Green's function for each spin
    '''

    energy_1body = energy_1body
    energy_2body = energy_2body
    fock_loop = fock_loop
    build_se_part = build_se_part

    def ao2mo(self, mo_coeff=None):
        ''' Get the electronic repulsion integrals in MO basis.
        '''

        nmo = max(self.nmo)
        mem_incore = ((nmo*(nmo+1)//2)**2) * 8/1e6
        mem_now = lib.current_memory()[0]

        if (self._scf._eri is not None and
                (mem_incore+mem_now < self.max_memory or self.incore_complete)):
            eri = _make_mo_eris_incore(self, mo_coeff)
        else:
            logger.warn(self, 'MO eris are outcore - this may be very '
                              'slow for agf2. increasing max_memory or '
                              'using density fitting is recommended.')
            eri = _make_mo_eris_outcore(self, mo_coeff)

        return eri

    def make_rdm1(self, gf=None):
        ''' Compute the one-body reduced density matrix in MO basis.

        Kwargs:
            gf : tuple of GreensFunction
                Auxiliaries of the Green's functions for each spin

        Returns:
            tuple of ndarray of density matrices
        '''

        if gf is None: gf = self.gf
        if gf is None: gf = self.init_gf()

        rdm1_a = gf[0].make_rdm1(occupancy=1)
        rdm1_b = gf[1].make_rdm1(occupancy=1)

        return (rdm1_a, rdm1_b)

    def get_fock(self, eri=None, gf=None, rdm1=None):
        ''' Computes the physical space Fock matrix in MO basis.
        '''

        if eri is None: eri = self.ao2mo()
        if gf is None: gf = self.gf

        return get_fock(self, eri, gf=gf, rdm1=rdm1)

    def energy_mp2(self, mo_energy=None, se=None):
        if mo_energy is None: mo_energy = self.mo_energy
        if se is None: se = self.build_se(gf=self.gf)

        self.e_init = energy_mp2(self, mo_energy, se)

        return self.e_init

    def init_gf(self, frozen=False):
        ''' Builds the Hartree-Fock Green's function.

        Returns:
            tuple of :class:`GreensFunction`, tuple of :class:`SelfEnergy`
        '''

        nmoa, nmob = self.nmo
        nocca, noccb = self.nocc

        energy = self.mo_energy
        coupling = (np.eye(nmoa), np.eye(nmob))

        focka = np.diag(energy[0])
        fockb = np.diag(energy[1])

        cpt_a = binsearch_chempot(focka, nmoa, nocca, occupancy=1)[0]
        cpt_b = binsearch_chempot(fockb, nmob, noccb, occupancy=1)[1]

        if frozen:
            mask = get_frozen_mask(self)
            energy = (energy[0][mask[0]], energy[1][mask[1]])
            coupling = (coupling[0][:,mask[0]], coupling[1][:,mask[1]])

        gf_a = aux.GreensFunction(energy[0], coupling[0], chempot=cpt_a)
        gf_b = aux.GreensFunction(energy[1], coupling[1], chempot=cpt_b)

        gf = (gf_a, gf_b)

        return gf

    def build_gf(self, eri=None, gf=None, se=None):
        ''' Builds the auxiliaries of the Green's functions by solving
            the Dyson equation for each spin.

        Kwargs:
            eri : _ChemistsERIs
                Electronic repulsion integrals
            gf : tuple of GreensFunction
                Auxiliaries of the Green's function for each spin
            se : tuple of SelfEnergy
                Auxiliaries of the self-energy for each spin

        Returns:
            tuple of :class:`GreensFunction`
        '''

        if eri is None: eri = self.ao2mo()
        if gf is None: gf = self.gf
        if gf is None: gf = self.init_gf()
        if se is None: se = self.build_se(eri, gf)

        focka, fockb = self.get_fock(eri, gf)

        gf_a = se[0].get_greens_function(focka)
        gf_b = se[1].get_greens_function(fockb)

        return (gf_a, gf_b)

    def build_se(self, eri=None, gf=None, os_factor=None, ss_factor=None, se_prev=None):
        ''' Builds the auxiliaries of the self-energy.

        Args:
            eri : _ChemistsERIs
                Electronic repulsion integrals
            gf : tuple of GreensFunction
                Auxiliaries of the Green's function

        Kwargs:
            os_factor : float
                Opposite-spin factor for spin-component-scaled (SCS)
                calculations. Default 1.0
            ss_factor : float
                Same-spin factor for spin-component-scaled (SCS)
                calculations. Default 1.0
            se_prev : SelfEnergy
                Previous self-energy for damping. Default value is None

        Returns
            tuple of :class:`SelfEnergy`
        '''

        if eri is None: eri = self.ao2mo()
        if gf is None: gf = self.gf
        if gf is None: gf = self.init_gf()

        if os_factor is None: os_factor = self.os_factor
        if ss_factor is None: ss_factor = self.ss_factor

        facs = {'os_factor': os_factor, 'ss_factor': ss_factor}
        gf_occ = (gf[0].get_occupied(), gf[1].get_occupied())
        gf_vir = (gf[0].get_virtual(), gf[1].get_virtual())

        se_occ = self.build_se_part(eri, gf_occ, gf_vir, **facs)
        se_vir = self.build_se_part(eri, gf_vir, gf_occ, **facs)

        se_a = aux.combine(se_occ[0], se_vir[0])
        se_b = aux.combine(se_occ[1], se_vir[1])

        if se_prev is not None and self.damping != 0.0:
            se_a_prev, se_b_prev = se_prev
            se_a.coupling *= np.sqrt(1.0-self.damping)
            se_b.coupling *= np.sqrt(1.0-self.damping)
            se_a_prev.coupling *= np.sqrt(self.damping)
            se_b_prev.coupling *= np.sqrt(self.damping)
            se_a = aux.combine(se_a, se_a_prev)
            se_b = aux.combine(se_b, se_b_prev)
            se_a = se_a.compress(n=(None,0))
            se_b = se_b.compress(n=(None,0))

        return (se_a, se_b)

    def run_diis(self, se, diis=None):
        ''' Runs the direct inversion of the iterative subspace for the
            self-energy.

        Args:
            se : SelfEnergy
                Auxiliaries of the self-energy
            diis : lib.diis.DIIS
                DIIS object

        Returns:
            tuple of :class:`SelfEnergy`
        '''

        if diis is None:
            return se

        se_occ_a, se_occ_b = (se[0].get_occupied(), se[1].get_occupied())
        se_vir_a, se_vir_b = (se[0].get_virtual(), se[1].get_virtual())

        vv_occ_a = np.dot(se_occ_a.coupling, se_occ_a.coupling.T)
        vv_occ_b = np.dot(se_occ_b.coupling, se_occ_b.coupling.T)
        vv_vir_a = np.dot(se_vir_a.coupling, se_vir_a.coupling.T)
        vv_vir_b = np.dot(se_vir_b.coupling, se_vir_b.coupling.T)

        vev_occ_a = np.dot(se_occ_a.coupling * se_occ_a.energy[None], se_occ_a.coupling.T)
        vev_occ_b = np.dot(se_occ_b.coupling * se_occ_b.energy[None], se_occ_b.coupling.T)
        vev_vir_a = np.dot(se_vir_a.coupling * se_vir_a.energy[None], se_vir_a.coupling.T)
        vev_vir_b = np.dot(se_vir_b.coupling * se_vir_b.energy[None], se_vir_b.coupling.T)

        dat = np.array([vv_occ_a, vv_vir_a, vev_occ_a, vev_vir_a,
                        vv_occ_b, vv_vir_b, vev_occ_b, vev_vir_b])
        dat = diis.update(dat)
        vv_occ_a, vv_vir_a, vev_occ_a, vev_vir_a, \
                vv_occ_b, vv_vir_b, vev_occ_b, vev_vir_b = dat

        se_occ_a = aux.SelfEnergy(*_agf2.cholesky_build(vv_occ_a, vev_occ_a), chempot=se[0].chempot)
        se_vir_a = aux.SelfEnergy(*_agf2.cholesky_build(vv_vir_a, vev_vir_a), chempot=se[0].chempot)
        se_occ_b = aux.SelfEnergy(*_agf2.cholesky_build(vv_occ_b, vev_occ_b), chempot=se[1].chempot)
        se_vir_b = aux.SelfEnergy(*_agf2.cholesky_build(vv_vir_b, vev_vir_b), chempot=se[1].chempot)
        se = (aux.combine(se_occ_a, se_vir_a), aux.combine(se_occ_b, se_vir_b))

        return se

    def density_fit(self, auxbasis=None, with_df=None):
        from pyscf.agf2 import dfuagf2
        myagf2 = dfuagf2.DFUAGF2(self._scf)
        myagf2.__dict__.update(self.__dict__)

        if with_df is not None:
            myagf2.with_df = with_df

        if auxbasis is not None and myagf2.with_df.auxbasis != auxbasis:
            myagf2.with_df = myagf2.with_df.copy()
            myagf2.with_df.auxbasis = auxbasis

        return myagf2

    def get_ip(self, gf, nroots=5):
        gf_occ = (gf[0].get_occupied(), gf[1].get_occupied())
        spin = np.array([0,]*gf_occ[0].naux + [1,]*gf_occ[1].naux)
        e_ip = np.concatenate([gf_occ[0].energy, gf_occ[1].energy], axis=0)
        v_ip = np.concatenate([gf_occ[0].coupling, gf_occ[1].coupling], axis=1)

        mask = np.argsort(e_ip)
        spin = list(spin[mask][-nroots:])[::-1]
        e_ip = list(-e_ip[mask][-nroots:])[::-1]
        v_ip = list(v_ip[:,mask][:,-nroots:].T)[::-1]

        return e_ip, v_ip, spin

    def ipagf2(self, nroots=5):
        e_ip, v_ip, spin = self.get_ip(self.gf, nroots=nroots)

        for n, en, vn, sn in zip(range(nroots), e_ip, v_ip, spin):
            qpwt = np.linalg.norm(vn)**2
            tag = ['alpha', 'beta'][sn]
            logger.note(self, 'IP energy level %d E = %.16g  QP weight = %0.6g  (%s)', n, en, qpwt, tag)

        if nroots == 1:
            return e_ip[0], v_ip[0]
        else:
            return e_ip, v_ip

    def get_ea(self, gf, nroots=5):
        gf_vir = (gf[0].get_virtual(), gf[1].get_virtual())
        spin = np.array([0,]*gf_vir[0].naux + [1,]*gf_vir[1].naux)
        e_ea = np.concatenate([gf_vir[0].energy, gf_vir[1].energy], axis=0)
        v_ea = np.concatenate([gf_vir[0].coupling, gf_vir[1].coupling], axis=1)

        mask = np.argsort(e_ea)
        spin = list(spin[mask][:nroots])
        e_ea = list(e_ea[mask][:nroots])
        v_ea = list(v_ea[:,mask][:,:nroots].T)

        return e_ea, v_ea, spin

    def eaagf2(self, nroots=5):
        e_ea, v_ea, spin = self.get_ea(self.gf, nroots=nroots)

        for n, en, vn, sn in zip(range(nroots), e_ea, v_ea, spin):
            qpwt = np.linalg.norm(vn)**2
            tag = ['alpha', 'beta'][sn]
            logger.note(self, 'EA energy level %d E = %.16g  QP weight = %0.6g  (%s)', n, en, qpwt, tag)

        if nroots == 1:
            return e_ea[0], v_ea[0]
        else:
            return e_ea, v_ea


    @property
    def nocc(self):
        if self._nocc is None:
            self._nocc = (np.sum(self.mo_occ[0] > 0), np.sum(self.mo_occ[1] > 0))
        return self._nocc
    @nocc.setter
    def nocc(self, val):
        self._nocc = val

    @property
    def nmo(self):
        if self._nmo is None:
            self._nmo = (self.mo_occ[0].size, self.mo_occ[1].size)
        return self._nmo
    @nmo.setter
    def nmo(self, val):
        self._nmo = val

    @property
    def qmo_energy(self):
        return (self.gf[0].energy, self.gf[1].energy)

    @property
    def qmo_coeff(self):
        ''' Gives the couplings in AO basis '''
        return (np.dot(self.mo_coeff[0], self.gf[0].coupling),
                np.dot(self.mo_coeff[1], self.gf[1].coupling))

    @property
    def qmo_occ(self):
        coeff_a = self.gf[0].get_occupied().coupling
        coeff_b = self.gf[1].get_occupied().coupling
        occ_a = np.linalg.norm(coeff_a, axis=0) ** 2
        occ_b = np.linalg.norm(coeff_b, axis=0) ** 2
        vir_a = np.zeros_like(self.gf[0].get_virtual().energy)
        vir_b = np.zeros_like(self.gf[1].get_virtual().energy)
        qmo_occ_a = np.concatenate([occ_a, vir_a])
        qmo_occ_b = np.concatenate([occ_b, vir_b])
        return qmo_occ_a, qmo_occ_b


def get_frozen_mask(agf2):
    with lib.temporary_env(agf2, _nocc=None, _nmo=None):
        return _get_frozen_mask(agf2)


class _ChemistsERIs:
    ''' (pq|rs)

    MO integrals stored in s4 symmetry, we only need QMO integrals
    in low-symmetry tensors and s4 is highest supported by _vhf
    '''

    def __init__(self, mol=None):
        self.mol = mol
        self.mo_coeff = None
        self.nocc = None
        self.nmo = None

        self.fock = None
        self.h1e = None
        self.eri = None
        self.e_hf = None

    def _common_init_(self, agf2, mo_coeff=None):
        if mo_coeff is None:
            mo_coeff = agf2.mo_coeff

        self.mo_coeff = mo_coeff

        dm = agf2._scf.make_rdm1(agf2.mo_coeff, agf2.mo_occ)
        h1e_ao = agf2._scf.get_hcore()
        vhf = agf2._scf.get_veff(agf2.mol, dm)
        fock_ao = agf2._scf.get_fock(vhf=vhf, dm=dm)

        self.h1e = (np.dot(np.dot(mo_coeff[0].conj().T, h1e_ao), mo_coeff[0]),
                    np.dot(np.dot(mo_coeff[1].conj().T, h1e_ao), mo_coeff[1]))
        self.fock = (np.dot(np.dot(mo_coeff[0].conj().T, fock_ao[0]), mo_coeff[0]),
                     np.dot(np.dot(mo_coeff[1].conj().T, fock_ao[1]), mo_coeff[1]))

        self.h1e = (mpi_helper.bcast(self.h1e[0]), mpi_helper.bcast(self.h1e[1]))
        self.fock = (mpi_helper.bcast(self.fock[0]), mpi_helper.bcast(self.fock[1]))

        self.e_hf = mpi_helper.bcast(agf2._scf.e_tot)

        self.nmo = agf2.nmo
        nocca, noccb = self.nocc = agf2.nocc
        self.mol = agf2.mol

        mo_e = (self.fock[0].diagonal(), self.fock[1].diagonal())
        gap_a = abs(mo_e[0][:nocca,None] - mo_e[0][None,nocca:]).min()
        gap_b = abs(mo_e[1][:noccb,None] - mo_e[1][None,noccb:]).min()
        gap = min(gap_a, gap_b)
        if gap < 1e-5:
            logger.warn(agf2, 'HOMO-LUMO gap %s may be too small for AGF2', gap)

        return self

def _make_mo_eris_incore(agf2, mo_coeff=None):
    ''' Returns _ChemistsERIs
    '''

    cput0 = (logger.process_clock(), logger.perf_counter())
    log = logger.Logger(agf2.stdout, agf2.verbose)

    eris = _ChemistsERIs()
    eris._common_init_(agf2, mo_coeff)
    moa, mob = eris.mo_coeff
    nmoa, nmob = eris.nmo

    eri_aa = ao2mo.incore.full(agf2._scf._eri, moa, verbose=log)
    eri_bb = ao2mo.incore.full(agf2._scf._eri, mob, verbose=log)

    eri_aa = ao2mo.addons.restore('s4', eri_aa, nmoa)
    eri_bb = ao2mo.addons.restore('s4', eri_bb, nmob)

    eri_ab = ao2mo.incore.general(agf2._scf._eri, (moa,moa,mob,mob), verbose=log)
    assert eri_ab.shape == (nmoa*(nmob+1)//2, nmob*(nmob+1)//2)
    eri_ba = np.transpose(eri_ab)

    eris.eri_aa = eri_aa
    eris.eri_ab = eri_ab
    eris.eri_ba = eri_ba
    eris.eri_bb = eri_bb
    eris.eri = ((eri_aa, eri_ab), (eri_ba, eri_bb))

    log.timer('MO integral transformation', *cput0)

    return eris

def _make_mo_eris_outcore(agf2, mo_coeff=None):
    ''' Returns _ChemistsERIs
    '''

    log = logger.Logger(agf2.stdout, agf2.verbose)

    eris = _ChemistsERIs()
    eris._common_init_(agf2, mo_coeff)

    mol = agf2.mol
    moa = np.asarray(eris.mo_coeff[0], order='F')
    mob = np.asarray(eris.mo_coeff[1], order='F')
    nmoa, nmob = eris.nmo

    eris.feri = lib.H5TmpFile()

    ao2mo.outcore.full(mol, moa, eris.feri, dataname='mo/aa')
    ao2mo.outcore.full(mol, mob, eris.feri, dataname='mo/bb')
    ao2mo.outcore.general(mol, (moa,moa,mob,mob), eris.feri, dataname='mo/ab', verbose=log)
    ao2mo.outcore.general(mol, (mob,mob,moa,moa), eris.feri, dataname='mo/ba', verbose=log)

    eris.eri_aa = eris.feri['mo/aa']
    eris.eri_ab = eris.feri['mo/ab']
    eris.eri_ba = eris.feri['mo/ba']
    eris.eri_bb = eris.feri['mo/bb']

    eris.eri = ((eris.eri_aa, eris.eri_ab), (eris.eri_ba, eris.eri_bb))

    return eris

def _make_qmo_eris_incore(agf2, eri, coeffs_a, coeffs_b, spin=None):
    ''' Returns nested tuple of ndarray

    spin = None: ((aaaa, aabb), (bbaa, bbbb))
    spin = 0: (aaaa, aabb)
    spin = 1: (bbbb, bbaa)
    '''

    cput0 = (logger.process_clock(), logger.perf_counter())
    log = logger.Logger(agf2.stdout, agf2.verbose)

    nmo = eri.nmo
    nmoa, nmob = nmo

    cxa = np.eye(nmoa)
    cxb = np.eye(nmob)
    if not (agf2.frozen is None or agf2.frozen == 0):
        mask = get_frozen_mask(agf2)
        cxa = cxa[:,mask[0]]
        cxb = cxb[:,mask[1]]

    # npaira, npairb = nmoa*(nmoa+1)//2, nmob*(nmob+1)//2
    cia, cja, caa = coeffs_a
    cib, cjb, cab = coeffs_b
    nia, nja, naa = [x.shape[1] for x in coeffs_a]
    nib, njb, nab = [x.shape[1] for x in coeffs_b]

    if spin is None or spin == 0:
        c_aa = (cxa, cia, cja, caa)
        c_ab = (cxa, cia, cjb, cab)

        qeri_aa = ao2mo.incore.general(eri.eri_aa, c_aa, compact=False, verbose=log)
        qeri_ab = ao2mo.incore.general(eri.eri_ab, c_ab, compact=False, verbose=log)

        qeri_aa = qeri_aa.reshape(cxa.shape[1], nia, nja, naa)
        qeri_ab = qeri_ab.reshape(cxa.shape[1], nia, njb, nab)

    if spin is None or spin == 1:
        c_bb = (cxb, cib, cjb, cab)
        c_ba = (cxb, cib, cja, caa)

        qeri_bb = ao2mo.incore.general(eri.eri_bb, c_bb, compact=False, verbose=log)
        qeri_ba = ao2mo.incore.general(eri.eri_ba, c_ba, compact=False, verbose=log)

        qeri_bb = qeri_bb.reshape(cxb.shape[1], nib, njb, nab)
        qeri_ba = qeri_ba.reshape(cxb.shape[1], nib, nja, naa)

    if spin is None:
        qeri = ((qeri_aa, qeri_ab), (qeri_ba, qeri_bb))
    elif spin == 0:
        qeri = (qeri_aa, qeri_ab)
    elif spin == 1:
        qeri = (qeri_bb, qeri_ba)

    log.timer('QMO integral transformation', *cput0)

    return qeri

def _make_qmo_eris_outcore(agf2, eri, coeffs_a, coeffs_b, spin=None):
    ''' Returns nested tuple of H5 dataset

    spin = None: ((aaaa, aabb), (bbaa, bbbb))
    spin = 0: (aaaa, aabb)
    spin = 1: (bbbb, bbaa)
    '''

    cput0 = (logger.process_clock(), logger.perf_counter())
    log = logger.Logger(agf2.stdout, agf2.verbose)

    nmo = eri.nmo
    nmoa, nmob = nmo

    mask = get_frozen_mask(agf2)
    frozena = np.sum(~mask[0])
    frozenb = np.sum(~mask[1])

    # npaira, npairb = nmoa*(nmoa+1)//2, nmob*(nmob+1)//2
    cia, cja, caa = coeffs_a
    cib, cjb, cab = coeffs_b
    nia, nja, naa = [x.shape[1] for x in coeffs_a]
    nib, njb, nab = [x.shape[1] for x in coeffs_b]

    # possible to have incore MO, outcore QMO
    if getattr(eri, 'feri', None) is None:
        eri.feri = lib.H5TmpFile()
    else:
        for key in ['aa', 'ab', 'ba', 'bb']:
            if 'qmo/%s'%key in eri.feri:
                del eri.feri['qmo/%s'%key]

    if spin is None or spin == 0:
        eri.feri.create_dataset('qmo/aa', (nmoa-frozena, nia, nja, naa), 'f8')
        eri.feri.create_dataset('qmo/ab', (nmoa-frozena, nia, njb, nab), 'f8')

        blksize = _agf2.get_blksize(agf2.max_memory, (nmoa**3, nmoa*nja*naa),
                                                (nmoa*nmob**2, nmoa*njb*nab))
        blksize = min(nmoa, max(BLKMIN, blksize))
        log.debug1('blksize (uagf2._make_qmo_eris_outcore) = %d', blksize)

        tril2sq = lib.square_mat_in_trilu_indices(nmoa)
        q1 = 0
        for p0, p1 in lib.prange(0, nmoa, blksize):
            if not np.any(mask[0][p0:p1]):
                # block is fully frozen
                continue

            inds = np.arange(p0, p1)[mask[0][p0:p1]]
            q0, q1 = q1, q1 + len(inds)
            idx = list(np.concatenate(tril2sq[inds]))

            # aa
            buf = eri.eri_aa[idx] # (blk, nmoa, npaira)
            buf = buf.reshape((q1-q0)*nmoa, -1) # (blk*nmoa, npaira)

            jasym_aa, nja_aa, cja_aa, sja_aa = ao2mo.incore._conc_mos(cja, caa)
            buf = ao2mo._ao2mo.nr_e2(buf, cja_aa, sja_aa, 's2kl', 's1')
            buf = buf.reshape(q1-q0, nmoa, nja, naa)

            buf = lib.einsum('xpja,pi->xija', buf, cia)
            eri.feri['qmo/aa'][q0:q1] = np.asarray(buf, order='C')

            # ab
            buf = eri.eri_ab[idx] # (blk, nmoa, npairb)
            buf = buf.reshape((q1-q0)*nmob, -1) # (blk*nmoa, npairb)

            jasym_ab, nja_ab, cja_ab, sja_ab = ao2mo.incore._conc_mos(cjb, cab)
            buf = ao2mo._ao2mo.nr_e2(buf, cja_ab, sja_ab, 's2kl', 's1')
            buf = buf.reshape(q1-q0, nmoa, njb, nab)

            buf = lib.einsum('xpja,pi->xija', buf, cia)
            eri.feri['qmo/ab'][q0:q1] = np.asarray(buf, order='C')

    if spin is None or spin == 1:
        eri.feri.create_dataset('qmo/ba', (nmob-frozenb, nib, nja, naa), 'f8')
        eri.feri.create_dataset('qmo/bb', (nmob-frozenb, nib, njb, nab), 'f8')

        max_memory = agf2.max_memory - lib.current_memory()[0]
        blksize = int((max_memory/8e-6) / max(nmob**3+nmob*njb*nab,
                                              nmob*nmoa**2*nja*naa))
        blksize = min(nmob, max(BLKMIN, blksize))
        log.debug1('blksize (uagf2._make_qmo_eris_outcore) = %d', blksize)

        tril2sq = lib.square_mat_in_trilu_indices(nmob)
        q1 = 0
        for p0, p1 in lib.prange(0, nmob, blksize):
            if not np.any(mask[1][p0:p1]):
                # block is fully frozen
                continue

            inds = np.arange(p0, p1)[mask[1][p0:p1]]
            q0, q1 = q1, q1 + len(inds)
            idx = list(np.concatenate(tril2sq[inds]))

            # ba
            buf = eri.eri_ba[idx] # (blk, nmob, npaira)
            buf = buf.reshape((q1-q0)*nmob, -1) # (blk*nmob, npaira)

            jasym_ba, nja_ba, cja_ba, sja_ba = ao2mo.incore._conc_mos(cja, caa)
            buf = ao2mo._ao2mo.nr_e2(buf, cja_ba, sja_ba, 's2kl', 's1')
            buf = buf.reshape(q1-q0, nmob, nja, naa)

            buf = lib.einsum('xpja,pi->xija', buf, cib)
            eri.feri['qmo/ba'][q0:q1] = np.asarray(buf, order='C')

            # bb
            buf = eri.eri_bb[idx] # (blk, nmob, npairb)
            buf = buf.reshape((q1-q0)*nmob, -1) # (blk*nmob, npairb)

            jasym_bb, nja_bb, cja_bb, sja_bb = ao2mo.incore._conc_mos(cjb, cab)
            buf = ao2mo._ao2mo.nr_e2(buf, cja_bb, sja_bb, 's2kl', 's1')
            buf = buf.reshape(q1-q0, nmob, njb, nab)

            buf = lib.einsum('xpja,pi->xija', buf, cib)
            eri.feri['qmo/bb'][q0:q1] = np.asarray(buf, order='C')

    if spin is None:
        qeri = ((eri.feri['qmo/aa'], eri.feri['qmo/ab']),
                (eri.feri['qmo/ba'], eri.feri['qmo/bb']))
    elif spin == 0:
        qeri = (eri.feri['qmo/aa'], eri.feri['qmo/ab'])
    elif spin == 1:
        qeri = (eri.feri['qmo/bb'], eri.feri['qmo/ba'])

    log.timer('QMO integral transformation', *cput0)

    return qeri



if __name__ == '__main__':
    from pyscf import gto, scf, mp

    mol = gto.M(atom='O 0 0 0; H 0 0 1; H 0 1 0', basis='cc-pvdz', charge=-1, spin=1, verbose=3)
    uhf = scf.UHF(mol)
    uhf.conv_tol = 1e-11
    uhf.run()

    uagf2 = UAGF2(uhf, frozen=0)

    uagf2.run()
    uagf2.ipagf2(nroots=5)
    uagf2.eaagf2(nroots=5)


    uagf2 = uagf2.density_fit()
    uagf2.run()



#!/usr/bin/env python
# Copyright 2014-2020 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from pyscf import lib
from pyscf import ao2mo
from pyscf.cc.rintermediates import _get_vvvv  # noqa
from pyscf.cc.ccsd import BLKMIN

# Ref: Gauss and Stanton, J. Chem. Phys. 103, 3561 (1995) Table III


def make_tau(t2, t1, r1, fac=1, out=None):
    t1a, t1b = t1
    r1a, r1b = r1
    tau1aa = make_tau_aa(t2[0], t1a, r1a, fac, out)
    tau1bb = make_tau_aa(t2[2], t1b, r1b, fac, out)
    tau1ab = make_tau_ab(t2[1], t1, r1, fac, out)
    return tau1aa, tau1ab, tau1bb

def make_tau_aa(t2aa, t1a, r1a, fac=1, out=None):
    tau1aa = np.einsum('ia,jb->ijab', t1a, r1a)
    tau1aa-= np.einsum('ia,jb->jiab', t1a, r1a)
    tau1aa = tau1aa - tau1aa.transpose(0,1,3,2)
    tau1aa *= fac * .5
    tau1aa += t2aa
    return tau1aa

def make_tau_ab(t2ab, t1, r1, fac=1, out=None):
    t1a, t1b = t1
    r1a, r1b = r1
    tau1ab = np.einsum('ia,jb->ijab', t1a, r1b)
    tau1ab+= np.einsum('ia,jb->ijab', r1a, t1b)
    tau1ab *= fac * .5
    tau1ab += t2ab
    return tau1ab

def Foo(t1, t2, eris):
    t1a, t1b = t1
    t2aa, t2ab, t2bb = t2
    nocca, noccb, nvira, nvirb = t2ab.shape

    eris_ovov = np.asarray(eris.ovov)
    eris_OVOV = np.asarray(eris.OVOV)
    eris_ovOV = np.asarray(eris.ovOV)
    tilaa, tilab, tilbb = make_tau(t2, t1, t1, fac=0.5)
    Fooa  = lib.einsum('inef,menf->mi', tilaa, eris_ovov)
    Fooa += lib.einsum('iNeF,meNF->mi', tilab, eris_ovOV)
    Foob  = lib.einsum('inef,menf->mi', tilbb, eris_OVOV)
    Foob += lib.einsum('nIfE,nfME->MI', tilab, eris_ovOV)

    eris_ovoo = np.asarray(eris.ovoo)
    eris_OVOO = np.asarray(eris.OVOO)
    eris_OVoo = np.asarray(eris.OVoo)
    eris_ovOO = np.asarray(eris.ovOO)
    Fooa += np.einsum('ne,nemi->mi', t1a, eris_ovoo)
    Fooa -= np.einsum('ne,meni->mi', t1a, eris_ovoo)
    Fooa += np.einsum('NE,NEmi->mi', t1b, eris_OVoo)
    Foob += np.einsum('ne,nemi->mi', t1b, eris_OVOO)
    Foob -= np.einsum('ne,meni->mi', t1b, eris_OVOO)
    Foob += np.einsum('ne,neMI->MI', t1a, eris_ovOO)

    fooa = eris.focka[:nocca,:nocca]
    foob = eris.fockb[:noccb,:noccb]
    fova = eris.focka[:nocca,nocca:]
    fovb = eris.fockb[:noccb,noccb:]
    Fova, Fovb = Fov(t1, t2, eris)

    Fooa += fooa + 0.5*lib.einsum('me,ie->mi', Fova+fova, t1a)
    Foob += foob + 0.5*lib.einsum('me,ie->mi', Fovb+fovb, t1b)
    return Fooa, Foob

def Fvv(t1, t2, eris):
    t1a, t1b = t1
    t2aa, t2ab, t2bb = t2
    nocca, noccb, nvira, nvirb = t2ab.shape

    Fvva = 0
    Fvvb = 0

    tauaa, tauab, taubb = make_tau(t2, t1, t1)
    #:eris_ovvv = lib.unpack_tril(np.asarray(eris.ovvv).reshape(nocca*nvira,-1)).reshape(nocca,nvira,nvira,nvira)
    #:ovvv = eris_ovvv - eris_ovvv.transpose(0,3,2,1)
    #:self.Fvva  = np.einsum('mf,mfae->ae', t1a, ovvv)
    mem_now = lib.current_memory()[0]
    max_memory = max(0, lib.param.MAX_MEMORY - mem_now)
    blksize = min(nocca, max(BLKMIN, int(max_memory*1e6/8/(nvira**3*3))))
    for p0,p1 in lib.prange(0, nocca, blksize):
        ovvv = eris.get_ovvv(slice(p0,p1))  # ovvv = eris.ovvv[p0:p1]
        ovvv = ovvv - ovvv.transpose(0,3,2,1)
        Fvva += np.einsum('mf,mfae->ae', t1a[p0:p1], ovvv)
        ovvv = None

    #:eris_OVVV = lib.unpack_tril(np.asarray(eris.OVVV).reshape(noccb*nvirb,-1)).reshape(noccb,nvirb,nvirb,nvirb)
    #:OVVV = eris_OVVV - eris_OVVV.transpose(0,3,2,1)
    #:self.Fvvb  = np.einsum('mf,mfae->ae', t1b, OVVV)
    #:self.wOVVO = lib.einsum('jf,mebf->mbej', t1b, OVVV)
    #:self.wOVOO  = 0.5 * lib.einsum('mebf,ijef->mbij', OVVV, taubb)
    blksize = min(noccb, max(BLKMIN, int(max_memory*1e6/8/(nvirb**3*3))))
    for p0, p1 in lib.prange(0, noccb, blksize):
        OVVV = eris.get_OVVV(slice(p0,p1))  # OVVV = eris.OVVV[p0:p1]
        OVVV = OVVV - OVVV.transpose(0,3,2,1)
        Fvvb += np.einsum('mf,mfae->ae', t1b[p0:p1], OVVV)
        OVVV = None

    #:eris_ovVV = lib.unpack_tril(np.asarray(eris.ovVV).reshape(nocca*nvira,-1)).reshape(nocca,nvira,nvirb,nvirb)
    #:self.Fvvb += np.einsum('mf,mfAE->AE', t1a, eris_ovVV)
    #:self.woVvO = lib.einsum('JF,meBF->mBeJ', t1b, eris_ovVV)
    #:self.woVVo = lib.einsum('jf,mfBE->mBEj',-t1a, eris_ovVV)
    #:self.woVoO  = 0.5 * lib.einsum('meBF,iJeF->mBiJ', eris_ovVV, tauab)
    #:self.woVoO += 0.5 * lib.einsum('mfBE,iJfE->mBiJ', eris_ovVV, tauab)
    blksize = min(nocca, max(BLKMIN, int(max_memory*1e6/8/(nvira*nvirb**2*3))))
    for p0,p1 in lib.prange(0, nocca, blksize):
        ovVV = eris.get_ovVV(slice(p0,p1))  # ovVV = eris.ovVV[p0:p1]
        Fvvb += np.einsum('mf,mfAE->AE', t1a[p0:p1], ovVV)
        ovVV = None

    #:eris_OVvv = lib.unpack_tril(np.asarray(eris.OVvv).reshape(noccb*nvirb,-1)).reshape(noccb,nvirb,nvira,nvira)
    #:self.Fvva += np.einsum('MF,MFae->ae', t1b, eris_OVvv)
    #:self.wOvVo = lib.einsum('jf,MEbf->MbEj', t1a, eris_OVvv)
    #:self.wOvvO = lib.einsum('JF,MFbe->MbeJ',-t1b, eris_OVvv)
    #:self.wOvOo  = 0.5 * lib.einsum('MEbf,jIfE->MbIj', eris_OVvv, tauab)
    #:self.wOvOo += 0.5 * lib.einsum('MFbe,jIeF->MbIj', eris_OVvv, tauab)
    blksize = min(noccb, max(BLKMIN, int(max_memory*1e6/8/(nvirb*nvira**2*3))))
    for p0, p1 in lib.prange(0, noccb, blksize):
        OVvv = eris.get_OVvv(slice(p0,p1))  # OVvv = eris.OVvv[p0:p1]
        Fvva += np.einsum('MF,MFae->ae', t1b[p0:p1], OVvv)
        OVvv = None

    eris_ovov = np.asarray(eris.ovov)
    eris_OVOV = np.asarray(eris.OVOV)
    eris_ovOV = np.asarray(eris.ovOV)
    tilaa, tilab, tilbb = make_tau(t2, t1, t1, fac=0.5)
    Fvva -= lib.einsum('mnaf,menf->ae', tilaa, eris_ovov)
    Fvva -= lib.einsum('mNaF,meNF->ae', tilab, eris_ovOV)
    Fvvb -= lib.einsum('mnaf,menf->ae', tilbb, eris_OVOV)
    Fvvb -= lib.einsum('nMfA,nfME->AE', tilab, eris_ovOV)

    fvva = eris.focka[nocca:,nocca:]
    fvvb = eris.fockb[noccb:,noccb:]
    fova = eris.focka[:nocca,nocca:]
    fovb = eris.fockb[:noccb,noccb:]
    Fova, Fovb = Fov(t1, t2, eris)

    Fvva += fvva - 0.5*lib.einsum('me,ma->ae', Fova+fova, t1a)
    Fvvb += fvvb - 0.5*lib.einsum('me,ma->ae', Fovb+fovb, t1b)
    return Fvva, Fvvb

def Fov(t1, t2, eris):
    t1a, t1b = t1
    t2aa, t2ab, t2bb = t2
    nocca, noccb, nvira, nvirb = t2ab.shape

    fova = eris.focka[:nocca,nocca:]
    fovb = eris.fockb[:noccb,noccb:]

    eris_ovov = np.asarray(eris.ovov)
    eris_OVOV = np.asarray(eris.OVOV)
    eris_ovOV = np.asarray(eris.ovOV)
    ovov = eris_ovov - eris_ovov.transpose(0,3,2,1)
    OVOV = eris_OVOV - eris_OVOV.transpose(0,3,2,1)

    Fova = np.einsum('nf,menf->me', t1a,      ovov)
    Fova+= np.einsum('NF,meNF->me', t1b, eris_ovOV)
    Fova += fova
    Fovb = np.einsum('nf,menf->me', t1b,      OVOV)
    Fovb+= np.einsum('nf,nfME->ME', t1a, eris_ovOV)
    Fovb += fovb
    return Fova, Fovb

def Woooo(t1, t2, eris):
    t1a, t1b = t1
    t2aa, t2ab, t2bb = t2
    eris_ovoo = np.asarray(eris.ovoo)
    eris_OVOO = np.asarray(eris.OVOO)
    eris_OVoo = np.asarray(eris.OVoo)
    eris_ovOO = np.asarray(eris.ovOO)
    ovoo = eris_ovoo - eris_ovoo.transpose(2,1,0,3)
    OVOO = eris_OVOO - eris_OVOO.transpose(2,1,0,3)
    woooo = lib.einsum('je,nemi->minj', t1a,      ovoo)
    wOOOO = lib.einsum('je,nemi->minj', t1b,      OVOO)
    wooOO = lib.einsum('JE,NEmi->miNJ', t1b, eris_OVoo)
    woOOo = lib.einsum('je,meNI->mINj',-t1a, eris_ovOO)

    woooo += np.asarray(eris.oooo)
    wOOOO += np.asarray(eris.OOOO)
    wooOO += np.asarray(eris.ooOO)
    woooo = woooo - woooo.transpose(0,3,2,1)
    wOOOO = wOOOO - wOOOO.transpose(0,3,2,1)
    wooOO = wooOO - woOOo.transpose(0,3,2,1)

    eris_ovov = np.asarray(eris.ovov)
    eris_OVOV = np.asarray(eris.OVOV)
    eris_ovOV = np.asarray(eris.ovOV)
    ovov = eris_ovov - eris_ovov.transpose(0,3,2,1)
    OVOV = eris_OVOV - eris_OVOV.transpose(0,3,2,1)
    tauaa, tauab, taubb = make_tau(t2, t1, t1)
    woooo += 0.5*lib.einsum('ijef,menf->minj', tauaa,      ovov)
    wOOOO += 0.5*lib.einsum('ijef,menf->minj', taubb,      OVOV)
    wooOO +=     lib.einsum('iJeF,meNF->miNJ', tauab, eris_ovOV)
    wOOoo = None
    return woooo, wooOO, wOOoo, wOOOO

def Wooov(t1, t2, eris):
    t1a, t1b = t1
    t2aa, t2ab, t2bb = t2
    dtype = np.result_type(t1a, t1b, t2aa, t2ab, t2bb)

    eris_ovoo = np.asarray(eris.ovoo)
    eris_OVOO = np.asarray(eris.OVOO)
    eris_OVoo = np.asarray(eris.OVoo)
    eris_ovOO = np.asarray(eris.ovOO)
    ovoo = eris_ovoo - eris_ovoo.transpose(2,1,0,3)
    OVOO = eris_OVOO - eris_OVOO.transpose(2,1,0,3)
    wooov = np.array(     ovoo.transpose(2,3,0,1), dtype=dtype)
    wOOOV = np.array(     OVOO.transpose(2,3,0,1), dtype=dtype)
    wooOV = np.array(eris_OVoo.transpose(2,3,0,1), dtype=dtype)
    wOOov = np.array(eris_ovOO.transpose(2,3,0,1), dtype=dtype)
    eris_ovoo = eris_OVOO = eris_ovOO = eris_OVoo = None

    eris_ovov = np.asarray(eris.ovov)
    eris_OVOV = np.asarray(eris.OVOV)
    eris_ovOV = np.asarray(eris.ovOV)
    ovov = eris_ovov - eris_ovov.transpose(0,3,2,1)
    OVOV = eris_OVOV - eris_OVOV.transpose(0,3,2,1)

    wooov += lib.einsum('if,mfne->mine', t1a,      ovov)
    wOOOV += lib.einsum('if,mfne->mine', t1b,      OVOV)
    wooOV += lib.einsum('if,mfNE->miNE', t1a, eris_ovOV)
    wOOov += lib.einsum('IF,neMF->MIne', t1b, eris_ovOV)
    return wooov, wooOV, wOOov, wOOOV

def Woovo(t1, t2, eris):
    t1a, t1b = t1
    t2aa, t2ab, t2bb = t2
    nocca, noccb, nvira, nvirb = t2ab.shape
    dtype = np.result_type(t1a, t1b, t2aa, t2ab, t2bb)

    wovoo = np.zeros((nocca,nvira,nocca,nocca), dtype=dtype)
    wOVOO = np.zeros((noccb,nvirb,noccb,noccb), dtype=dtype)
    woVoO = np.zeros((nocca,nvirb,nocca,noccb), dtype=dtype)
    wOvOo = np.zeros((noccb,nvira,noccb,nocca), dtype=dtype)

    tauaa, tauab, taubb = make_tau(t2, t1, t1)
    #:eris_ovvv = lib.unpack_tril(np.asarray(eris.ovvv).reshape(nocca*nvira,-1)).reshape(nocca,nvira,nvira,nvira)
    #:ovvv = eris_ovvv - eris_ovvv.transpose(0,3,2,1)
    #:self.wovoo  = 0.5 * lib.einsum('mebf,ijef->mbij', eris_ovvv, tauaa)
    #:self.wovoo -= 0.5 * lib.einsum('mfbe,ijef->mbij', eris_ovvv, tauaa)
    mem_now = lib.current_memory()[0]
    max_memory = max(0, lib.param.MAX_MEMORY - mem_now)
    blksize = min(nocca, max(BLKMIN, int(max_memory*1e6/8/(nvira**3*3))))
    for p0,p1 in lib.prange(0, nocca, blksize):
        ovvv = eris.get_ovvv(slice(p0,p1))  # ovvv = eris.ovvv[p0:p1]
        ovvv = ovvv - ovvv.transpose(0,3,2,1)
        wovoo[p0:p1] = 0.5 * lib.einsum('mebf,ijef->mbij', ovvv, tauaa)
        ovvv = None

    #:eris_OVVV = lib.unpack_tril(np.asarray(eris.OVVV).reshape(noccb*nvirb,-1)).reshape(noccb,nvirb,nvirb,nvirb)
    #:OVVV = eris_OVVV - eris_OVVV.transpose(0,3,2,1)
    #:self.wOVOO  = 0.5 * lib.einsum('mebf,ijef->mbij', OVVV, taubb)
    blksize = min(noccb, max(BLKMIN, int(max_memory*1e6/8/(nvirb**3*3))))
    for p0, p1 in lib.prange(0, noccb, blksize):
        OVVV = eris.get_OVVV(slice(p0,p1))  # OVVV = eris.OVVV[p0:p1]
        OVVV = OVVV - OVVV.transpose(0,3,2,1)
        wOVOO[p0:p1] = 0.5 * lib.einsum('mebf,ijef->mbij', OVVV, taubb)
        OVVV = None

    #:eris_ovVV = lib.unpack_tril(np.asarray(eris.ovVV).reshape(nocca*nvira,-1)).reshape(nocca,nvira,nvirb,nvirb)
    #:self.Fvvb += np.einsum('mf,mfAE->AE', t1a, eris_ovVV)
    #:self.woVvO = lib.einsum('JF,meBF->mBeJ', t1b, eris_ovVV)
    #:self.woVVo = lib.einsum('jf,mfBE->mBEj',-t1a, eris_ovVV)
    #:self.woVoO  = 0.5 * lib.einsum('meBF,iJeF->mBiJ', eris_ovVV, tauab)
    #:self.woVoO += 0.5 * lib.einsum('mfBE,iJfE->mBiJ', eris_ovVV, tauab)
    blksize = min(nocca, max(BLKMIN, int(max_memory*1e6/8/(nvira*nvirb**2*3))))
    for p0,p1 in lib.prange(0, nocca, blksize):
        ovVV = eris.get_ovVV(slice(p0,p1))  # ovVV = eris.ovVV[p0:p1]
        woVoO[p0:p1] = 0.5 * lib.einsum('meBF,iJeF->mBiJ', ovVV, tauab)
        woVoO[p0:p1]+= 0.5 * lib.einsum('mfBE,iJfE->mBiJ', ovVV, tauab)
        ovVV = None

    #:eris_OVvv = lib.unpack_tril(np.asarray(eris.OVvv).reshape(noccb*nvirb,-1)).reshape(noccb,nvirb,nvira,nvira)
    #:self.Fvva += np.einsum('MF,MFae->ae', t1b, eris_OVvv)
    #:self.wOvVo = lib.einsum('jf,MEbf->MbEj', t1a, eris_OVvv)
    #:self.wOvvO = lib.einsum('JF,MFbe->MbeJ',-t1b, eris_OVvv)
    #:self.wOvOo  = 0.5 * lib.einsum('MEbf,jIfE->MbIj', eris_OVvv, tauab)
    #:self.wOvOo += 0.5 * lib.einsum('MFbe,jIeF->MbIj', eris_OVvv, tauab)
    blksize = min(noccb, max(BLKMIN, int(max_memory*1e6/8/(nvirb*nvira**2*3))))
    for p0, p1 in lib.prange(0, noccb, blksize):
        OVvv = eris.get_OVvv(slice(p0,p1))  # OVvv = eris.OVvv[p0:p1]
        wOvOo[p0:p1] = 0.5 * lib.einsum('MEbf,jIfE->MbIj', OVvv, tauab)
        wOvOo[p0:p1]+= 0.5 * lib.einsum('MFbe,jIeF->MbIj', OVvv, tauab)
        OVvv = None

    eris_ovov = np.asarray(eris.ovov)
    eris_OVOV = np.asarray(eris.OVOV)
    eris_ovOV = np.asarray(eris.ovOV)
    ovov = eris_ovov - eris_ovov.transpose(0,3,2,1)
    OVOV = eris_OVOV - eris_OVOV.transpose(0,3,2,1)
    tilaa, tilab, tilbb = make_tau(t2, t1, t1, fac=0.5)

    eris_ovoo = np.asarray(eris.ovoo)
    eris_OVOO = np.asarray(eris.OVOO)
    eris_OVoo = np.asarray(eris.OVoo)
    eris_ovOO = np.asarray(eris.ovOO)
    ovoo = eris_ovoo - eris_ovoo.transpose(2,1,0,3)
    OVOO = eris_OVOO - eris_OVOO.transpose(2,1,0,3)
    tmpaa = lib.einsum('nemi,jnbe->mbij',      ovoo, t2aa)
    tmpaa+= lib.einsum('NEmi,jNbE->mbij', eris_OVoo, t2ab)
    tmpbb = lib.einsum('nemi,jnbe->mbij',      OVOO, t2bb)
    tmpbb+= lib.einsum('neMI,nJeB->MBIJ', eris_ovOO, t2ab)
    woVoO += lib.einsum('nemi,nJeB->mBiJ',      ovoo, t2ab)
    woVoO += lib.einsum('NEmi,JNBE->mBiJ', eris_OVoo, t2bb)
    woVoO -= lib.einsum('meNI,jNeB->mBjI', eris_ovOO, t2ab)
    wOvOo += lib.einsum('NEMI,jNbE->MbIj',      OVOO, t2ab)
    wOvOo += lib.einsum('neMI,jnbe->MbIj', eris_ovOO, t2aa)
    wOvOo -= lib.einsum('MEni,nJbE->MbJi', eris_OVoo, t2ab)
    wovoo += tmpaa - tmpaa.transpose(0,1,3,2)
    wOVOO += tmpbb - tmpbb.transpose(0,1,3,2)

    tmp1aa = lib.einsum('njbf,menf->mbej', t2aa,      ovov)
    tmp1aa-= lib.einsum('jNbF,meNF->mbej', t2ab, eris_ovOV)
    tmp1bb = lib.einsum('njbf,menf->mbej', t2bb,      OVOV)
    tmp1bb-= lib.einsum('nJfB,nfME->MBEJ', t2ab, eris_ovOV)
    tmp1ab = lib.einsum('NJBF,meNF->mBeJ', t2bb, eris_ovOV)
    tmp1ab-= lib.einsum('nJfB,menf->mBeJ', t2ab,      ovov)
    tmp1ba = lib.einsum('njbf,nfME->MbEj', t2aa, eris_ovOV)
    tmp1ba-= lib.einsum('jNbF,MENF->MbEj', t2ab,      OVOV)
    tmp1abba =-lib.einsum('jNfB,mfNE->mBEj', t2ab, eris_ovOV)
    tmp1baab =-lib.einsum('nJbF,neMF->MbeJ', t2ab, eris_ovOV)
    tmpaa = lib.einsum('ie,mbej->mbij', t1a, tmp1aa)
    tmpbb = lib.einsum('ie,mbej->mbij', t1b, tmp1bb)
    tmpab = lib.einsum('ie,mBeJ->mBiJ', t1a, tmp1ab)
    tmpab-= lib.einsum('IE,mBEj->mBjI', t1b, tmp1abba)
    tmpba = lib.einsum('IE,MbEj->MbIj', t1b, tmp1ba)
    tmpba-= lib.einsum('ie,MbeJ->MbJi', t1a, tmp1baab)
    wovoo -= tmpaa - tmpaa.transpose(0,1,3,2)
    wOVOO -= tmpbb - tmpbb.transpose(0,1,3,2)
    woVoO -= tmpab
    wOvOo -= tmpba
    eris_ovov = eris_OVOV = eris_ovOV = None
    eris_ovoo = np.asarray(eris.ovoo)
    eris_OVOO = np.asarray(eris.OVOO)
    eris_ovOO = np.asarray(eris.ovOO)
    eris_OVoo = np.asarray(eris.OVoo)
    wovoo += eris_ovoo.transpose(3,1,2,0) - eris_ovoo.transpose(2,1,0,3)
    wOVOO += eris_OVOO.transpose(3,1,2,0) - eris_OVOO.transpose(2,1,0,3)
    woVoO += eris_OVoo.transpose(3,1,2,0)
    wOvOo += eris_ovOO.transpose(3,1,2,0)
    eris_ovoo = eris_OVOO = eris_ovOO = eris_OVoo = None

    eris_ovvo = np.asarray(eris.ovvo)
    eris_OVVO = np.asarray(eris.OVVO)
    eris_OVvo = np.asarray(eris.OVvo)
    eris_ovVO = np.asarray(eris.ovVO)
    eris_oovv = np.asarray(eris.oovv)
    eris_OOVV = np.asarray(eris.OOVV)
    eris_OOvv = np.asarray(eris.OOvv)
    eris_ooVV = np.asarray(eris.ooVV)
    tmpaa = lib.einsum('ie,mebj->mbij', t1a, eris_ovvo)
    tmpbb = lib.einsum('ie,mebj->mbij', t1b, eris_OVVO)
    tmpaa-= lib.einsum('ie,mjbe->mbij', t1a, eris_oovv)
    tmpbb-= lib.einsum('ie,mjbe->mbij', t1b, eris_OOVV)
    woVoO += lib.einsum('ie,meBJ->mBiJ', t1a, eris_ovVO)
    woVoO -= lib.einsum('IE,mjBE->mBjI',-t1b, eris_ooVV)
    wOvOo += lib.einsum('IE,MEbj->MbIj', t1b, eris_OVvo)
    wOvOo -= lib.einsum('ie,MJbe->MbJi',-t1a, eris_OOvv)
    wovoo += tmpaa - tmpaa.transpose(0,1,3,2)
    wOVOO += tmpbb - tmpbb.transpose(0,1,3,2)

    woooo, wooOO, wOOoo, wOOOO = Woooo(t1, t2, eris)
    Fova, Fovb = Fov(t1, t2, eris)
    wovoo -= lib.einsum('me,ijbe->mbij', Fova, t2aa)
    wOVOO -= lib.einsum('me,ijbe->mbij', Fovb, t2bb)
    woVoO += lib.einsum('me,iJeB->mBiJ', Fova, t2ab)
    wOvOo += lib.einsum('ME,jIbE->MbIj', Fovb, t2ab)
    wovoo -= lib.einsum('nb,minj->mbij', t1a, woooo)
    wOVOO -= lib.einsum('nb,minj->mbij', t1b, wOOOO)
    woVoO -= lib.einsum('NB,miNJ->mBiJ', t1b, wooOO)
    wOvOo -= lib.einsum('nb,njMI->MbIj', t1a, wooOO)
    eris_ovvo = eris_OVVO = eris_OVvo = eris_ovVO = None
    eris_oovv = eris_OOVV = eris_OOvv = eris_ooVV = None

    woovo = wovoo.transpose(0,2,1,3)
    wOOVO = wOVOO.transpose(0,2,1,3)
    wooVO = woVoO.transpose(0,2,1,3)
    wOOvo = wOvOo.transpose(0,2,1,3)
    return woovo, wooVO, wOOvo, wOOVO

def Wovvo(t1, t2, eris):
    t1a, t1b = t1
    t2aa, t2ab, t2bb = t2
    nocca, noccb, nvira, nvirb = t2ab.shape
    dtype = np.result_type(t1a, t1b, t2aa, t2ab, t2bb)

    wovvo = np.zeros((nocca,nvira,nvira,nocca), dtype=dtype)
    wOVVO = np.zeros((noccb,nvirb,nvirb,noccb), dtype=dtype)
    woVvO = np.zeros((nocca,nvirb,nvira,noccb), dtype=dtype)
    woVVo = np.zeros((nocca,nvirb,nvirb,nocca), dtype=dtype)
    wOvVo = np.zeros((noccb,nvira,nvirb,nocca), dtype=dtype)
    wOvvO = np.zeros((noccb,nvira,nvira,noccb), dtype=dtype)

    tauaa, tauab, taubb = make_tau(t2, t1, t1)
    #:eris_ovvv = lib.unpack_tril(np.asarray(eris.ovvv).reshape(nocca*nvira,-1)).reshape(nocca,nvira,nvira,nvira)
    #:ovvv = eris_ovvv - eris_ovvv.transpose(0,3,2,1)
    #:self.Fvva  = np.einsum('mf,mfae->ae', t1a, ovvv)
    #:self.wovvo = lib.einsum('jf,mebf->mbej', t1a, ovvv)
    #:self.wovoo  = 0.5 * lib.einsum('mebf,ijef->mbij', eris_ovvv, tauaa)
    #:self.wovoo -= 0.5 * lib.einsum('mfbe,ijef->mbij', eris_ovvv, tauaa)
    mem_now = lib.current_memory()[0]
    max_memory = max(0, lib.param.MAX_MEMORY - mem_now)
    blksize = min(nocca, max(BLKMIN, int(max_memory*1e6/8/(nvira**3*3))))
    for p0,p1 in lib.prange(0, nocca, blksize):
        ovvv = eris.get_ovvv(slice(p0,p1))  # ovvv = eris.ovvv[p0:p1]
        ovvv = ovvv - ovvv.transpose(0,3,2,1)
        wovvo[p0:p1] = lib.einsum('jf,mebf->mbej', t1a, ovvv)
        ovvv = None

    #:eris_OVVV = lib.unpack_tril(np.asarray(eris.OVVV).reshape(noccb*nvirb,-1)).reshape(noccb,nvirb,nvirb,nvirb)
    #:OVVV = eris_OVVV - eris_OVVV.transpose(0,3,2,1)
    #:self.Fvvb  = np.einsum('mf,mfae->ae', t1b, OVVV)
    #:self.wOVVO = lib.einsum('jf,mebf->mbej', t1b, OVVV)
    #:self.wOVOO  = 0.5 * lib.einsum('mebf,ijef->mbij', OVVV, taubb)
    blksize = min(noccb, max(BLKMIN, int(max_memory*1e6/8/(nvirb**3*3))))
    for p0, p1 in lib.prange(0, noccb, blksize):
        OVVV = eris.get_OVVV(slice(p0,p1))  # OVVV = eris.OVVV[p0:p1]
        OVVV = OVVV - OVVV.transpose(0,3,2,1)
        wOVVO[p0:p1] = lib.einsum('jf,mebf->mbej', t1b, OVVV)
        OVVV = None

    #:eris_ovVV = lib.unpack_tril(np.asarray(eris.ovVV).reshape(nocca*nvira,-1)).reshape(nocca,nvira,nvirb,nvirb)
    #:self.Fvvb += np.einsum('mf,mfAE->AE', t1a, eris_ovVV)
    #:self.woVvO = lib.einsum('JF,meBF->mBeJ', t1b, eris_ovVV)
    #:self.woVVo = lib.einsum('jf,mfBE->mBEj',-t1a, eris_ovVV)
    #:self.woVoO  = 0.5 * lib.einsum('meBF,iJeF->mBiJ', eris_ovVV, tauab)
    #:self.woVoO += 0.5 * lib.einsum('mfBE,iJfE->mBiJ', eris_ovVV, tauab)
    blksize = min(nocca, max(BLKMIN, int(max_memory*1e6/8/(nvira*nvirb**2*3))))
    for p0,p1 in lib.prange(0, nocca, blksize):
        ovVV = eris.get_ovVV(slice(p0,p1))  # ovVV = eris.ovVV[p0:p1]
        woVvO[p0:p1] = lib.einsum('JF,meBF->mBeJ', t1b, ovVV)
        woVVo[p0:p1] = lib.einsum('jf,mfBE->mBEj',-t1a, ovVV)
        ovVV = None

    #:eris_OVvv = lib.unpack_tril(np.asarray(eris.OVvv).reshape(noccb*nvirb,-1)).reshape(noccb,nvirb,nvira,nvira)
    #:self.Fvva += np.einsum('MF,MFae->ae', t1b, eris_OVvv)
    #:self.wOvVo = lib.einsum('jf,MEbf->MbEj', t1a, eris_OVvv)
    #:self.wOvvO = lib.einsum('JF,MFbe->MbeJ',-t1b, eris_OVvv)
    #:self.wOvOo  = 0.5 * lib.einsum('MEbf,jIfE->MbIj', eris_OVvv, tauab)
    #:self.wOvOo += 0.5 * lib.einsum('MFbe,jIeF->MbIj', eris_OVvv, tauab)
    blksize = min(noccb, max(BLKMIN, int(max_memory*1e6/8/(nvirb*nvira**2*3))))
    for p0, p1 in lib.prange(0, noccb, blksize):
        OVvv = eris.get_OVvv(slice(p0,p1))  # OVvv = eris.OVvv[p0:p1]
        wOvVo[p0:p1] = lib.einsum('jf,MEbf->MbEj', t1a, OVvv)
        wOvvO[p0:p1] = lib.einsum('JF,MFbe->MbeJ',-t1b, OVvv)
        OVvv = None

    eris_ovov = np.asarray(eris.ovov)
    eris_OVOV = np.asarray(eris.OVOV)
    eris_ovOV = np.asarray(eris.ovOV)
    ovov = eris_ovov - eris_ovov.transpose(0,3,2,1)
    OVOV = eris_OVOV - eris_OVOV.transpose(0,3,2,1)
    wovvo -= lib.einsum('jnfb,menf->mbej', t2aa,      ovov)
    wovvo += lib.einsum('jNbF,meNF->mbej', t2ab, eris_ovOV)
    wOVVO -= lib.einsum('jnfb,menf->mbej', t2bb,      OVOV)
    wOVVO += lib.einsum('nJfB,nfME->MBEJ', t2ab, eris_ovOV)
    woVvO += lib.einsum('nJfB,menf->mBeJ', t2ab,      ovov)
    woVvO -= lib.einsum('JNFB,meNF->mBeJ', t2bb, eris_ovOV)
    wOvVo -= lib.einsum('jnfb,nfME->MbEj', t2aa, eris_ovOV)
    wOvVo += lib.einsum('jNbF,MENF->MbEj', t2ab,      OVOV)
    woVVo += lib.einsum('jNfB,mfNE->mBEj', t2ab, eris_ovOV)
    wOvvO += lib.einsum('nJbF,neMF->MbeJ', t2ab, eris_ovOV)

    eris_ovoo = np.asarray(eris.ovoo)
    eris_OVOO = np.asarray(eris.OVOO)
    eris_OVoo = np.asarray(eris.OVoo)
    eris_ovOO = np.asarray(eris.ovOO)
    eris_ovoo = eris_ovoo + np.einsum('nfme,jf->menj', eris_ovov, t1a)
    eris_OVOO = eris_OVOO + np.einsum('nfme,jf->menj', eris_OVOV, t1b)
    eris_OVoo = eris_OVoo + np.einsum('nfme,jf->menj', eris_ovOV, t1a)
    eris_ovOO = eris_ovOO + np.einsum('menf,jf->menj', eris_ovOV, t1b)
    ovoo = eris_ovoo - eris_ovoo.transpose(2,1,0,3)
    OVOO = eris_OVOO - eris_OVOO.transpose(2,1,0,3)
    wovvo += lib.einsum('nb,nemj->mbej', t1a,      ovoo)
    wOVVO += lib.einsum('nb,nemj->mbej', t1b,      OVOO)
    woVvO -= lib.einsum('NB,meNJ->mBeJ', t1b, eris_ovOO)
    wOvVo -= lib.einsum('nb,MEnj->MbEj', t1a, eris_OVoo)
    woVVo += lib.einsum('NB,NEmj->mBEj', t1b, eris_OVoo)
    wOvvO += lib.einsum('nb,neMJ->MbeJ', t1a, eris_ovOO)
    eris_ovoo = eris_OVOO = eris_ovOO = eris_OVoo = None

    eris_ovvo = np.asarray(eris.ovvo)
    eris_OVVO = np.asarray(eris.OVVO)
    eris_OVvo = np.asarray(eris.OVvo)
    eris_ovVO = np.asarray(eris.ovVO)
    eris_oovv = np.asarray(eris.oovv)
    eris_OOVV = np.asarray(eris.OOVV)
    eris_OOvv = np.asarray(eris.OOvv)
    eris_ooVV = np.asarray(eris.ooVV)

    wovvo = wovvo.transpose(0,2,1,3)
    wOVVO = wOVVO.transpose(0,2,1,3)
    wovVO = woVvO.transpose(0,2,1,3)
    wOVvo = wOvVo.transpose(0,2,1,3)
    woVVo = woVVo.transpose(0,2,1,3)
    wOvvO = wOvvO.transpose(0,2,1,3)

    wovvo += eris_ovvo
    wOVVO += eris_OVVO
    wovVO += eris_ovVO
    wOVvo += eris_OVvo
    wovvo -= eris_oovv.transpose(0,3,2,1)
    wOVVO -= eris_OOVV.transpose(0,3,2,1)
    woVVo -= eris_ooVV.transpose(0,3,2,1)
    wOvvO -= eris_OOvv.transpose(0,3,2,1)
    return wovvo, wovVO, wOVvo, wOVVO, woVVo, wOvvO

def Wvvov(t1, t2, eris):
    t1a, t1b = t1
    t2aa, t2ab, t2bb = t2
    nocca, noccb, nvira, nvirb = t2ab.shape

    #:Wamef = einsum('na,nmef->amef', -t1, eris.oovv)
    #:Wamef -= np.asarray(eris.ovvv).transpose(1,0,2,3)
    eris_ovov = np.asarray(eris.ovov)
    eris_OVOV = np.asarray(eris.OVOV)
    eris_ovOV = np.asarray(eris.ovOV)
    Waemf = lib.einsum('na,nemf->aemf',-t1a, eris_ovov)
    Waemf+= np.asarray(eris.get_ovvv()).transpose(2,3,0,1)
    Waemf = Waemf - Waemf.transpose(0,3,2,1)

    WaeMF = lib.einsum('na,nemf->aemf',-t1a, eris_ovOV)
    WaeMF+= np.asarray(eris.get_OVvv()).transpose(2,3,0,1)

    WAEmf = lib.einsum('na,mfne->aemf',-t1b, eris_ovOV)
    WAEmf+= np.asarray(eris.get_ovVV()).transpose(2,3,0,1)

    WAEMF = lib.einsum('na,nemf->aemf',-t1b, eris_OVOV)
    WAEMF+= np.asarray(eris.get_OVVV()).transpose(2,3,0,1)
    WAEMF = WAEMF - WAEMF.transpose(0,3,2,1)
    return Waemf, WaeMF, WAEmf, WAEMF

def Wvvvo(t1, t2, eris):
    t1a, t1b = t1
    t2aa, t2ab, t2bb = t2
    nocca, noccb, nvira, nvirb = t2ab.shape
    dtype = np.result_type(t1a, t1b, t2aa, t2ab, t2bb)

    _wvovv = np.empty((nvira,nocca,nvira,nvira), dtype=dtype)
    _wVOVV = np.empty((nvirb,noccb,nvirb,nvirb), dtype=dtype)
    _wvOvV = np.empty((nvira,noccb,nvira,nvirb), dtype=dtype)
    _wVoVv = np.empty((nvirb,nocca,nvirb,nvira), dtype=dtype)

    Fova, Fovb = Fov(t1, t2, eris)

    # 3 or 4 virtuals
    eris_ovoo = np.asarray(eris.ovoo)
    eris_ovov = np.asarray(eris.ovov)
    eris_ovOV = np.asarray(eris.ovOV)
    ovov = eris_ovov - eris_ovov.transpose(0,3,2,1)
    eris_oovv = np.asarray(eris.oovv)
    eris_ovvo = np.asarray(eris.ovvo)
    oovv = eris_oovv - eris_ovvo.transpose(0,3,2,1)
    eris_oovv = eris_ovvo = None
    tauaa, tauab, taubb = make_tau(t2, t1, t1)
    #:wvovv  = .5 * lib.einsum('meni,mnab->eiab', eris_ovoo, tauaa)
    #:wvovv -= .5 * lib.einsum('me,miab->eiab', self.Fova, t2aa)
    #:tmp1aa = lib.einsum('nibf,menf->mbei', t2aa,      ovov)
    #:tmp1aa-= lib.einsum('iNbF,meNF->mbei', t2ab, eris_ovOV)
    #:wvovv+= lib.einsum('ma,mbei->eiab', t1a, tmp1aa)
    #:wvovv+= lib.einsum('ma,mibe->eiab', t1a,      oovv)
    for p0, p1 in lib.prange(0, nvira, nocca):
        wvovv  = .5*lib.einsum('meni,mnab->eiab', eris_ovoo[:,p0:p1], tauaa)
        wvovv -= .5*lib.einsum('me,miab->eiab', Fova[:,p0:p1], t2aa)

        tmp1aa = lib.einsum('nibf,menf->mbei', t2aa, ovov[:,p0:p1])
        tmp1aa-= lib.einsum('iNbF,meNF->mbei', t2ab, eris_ovOV[:,p0:p1])
        wvovv += lib.einsum('ma,mbei->eiab', t1a, tmp1aa)
        wvovv += lib.einsum('ma,mibe->eiab', t1a, oovv[:,:,:,p0:p1])
        _wvovv[p0:p1] = wvovv
        tmp1aa = None
    eris_ovov = eris_ovoo = eris_ovOV = None

    #:eris_ovvv = lib.unpack_tril(np.asarray(eris.ovvv).reshape(nocca*nvira,-1)).reshape(nocca,nvira,nvira,nvira)
    #:ovvv = eris_ovvv - eris_ovvv.transpose(0,3,2,1)
    #:wvovv += lib.einsum('mebf,miaf->eiab',      ovvv, t2aa)
    #:eris_OVvv = lib.unpack_tril(np.asarray(eris.OVvv).reshape(noccb*nvirb,-1)).reshape(noccb,nvirb,nvira,nvira)
    #:wvovv += lib.einsum('MFbe,iMaF->eiab', eris_OVvv, t2ab)
    #:wvovv += eris_ovvv.transpose(2,0,3,1).conj()
    #:self.wvovv -= wvovv - wvovv.transpose(0,1,3,2)
    mem_now = lib.current_memory()[0]
    max_memory = max(0, lib.param.MAX_MEMORY - mem_now)
    blksize = min(nocca, max(BLKMIN, int(max_memory*1e6/8/(nvira**3*6))))
    for i0,i1 in lib.prange(0, nocca, blksize):
        wvovv = _wvovv[:,i0:i1]
        for p0,p1 in lib.prange(0, noccb, blksize):
            OVvv = eris.get_OVvv(slice(p0,p1))  # OVvv = eris.OVvv[p0:p1]
            wvovv -= lib.einsum('MFbe,iMaF->eiab', OVvv, t2ab[i0:i1,p0:p1])
            OVvv = None
        for p0,p1 in lib.prange(0, nocca, blksize):
            ovvv = eris.get_ovvv(slice(p0,p1))  # ovvv = eris.ovvv[p0:p1]
            if p0 == i0:
                wvovv += ovvv.transpose(2,0,3,1).conj()
            ovvv = ovvv - ovvv.transpose(0,3,2,1)
            wvovv -= lib.einsum('mebf,miaf->eiab', ovvv, t2aa[p0:p1,i0:i1])
            ovvv = None
        wvovv = wvovv - wvovv.transpose(0,1,3,2)
        _wvovv[:,i0:i1] = wvovv

    eris_OVOO = np.asarray(eris.OVOO)
    eris_OVOV = np.asarray(eris.OVOV)
    eris_ovOV = np.asarray(eris.ovOV)
    OVOV = eris_OVOV - eris_OVOV.transpose(0,3,2,1)
    eris_OOVV = np.asarray(eris.OOVV)
    eris_OVVO = np.asarray(eris.OVVO)
    OOVV = eris_OOVV - eris_OVVO.transpose(0,3,2,1)
    eris_OOVV = eris_OVVO = None
    #:wVOVV  = .5*lib.einsum('meni,mnab->eiab', eris_OVOO, taubb)
    #:wVOVV -= .5*lib.einsum('me,miab->eiab', self.Fovb, t2bb)
    #:tmp1bb = lib.einsum('nibf,menf->mbei', t2bb,      OVOV)
    #:tmp1bb-= lib.einsum('nIfB,nfME->MBEI', t2ab, eris_ovOV)
    #:wVOVV += lib.einsum('ma,mbei->eiab', t1b, tmp1bb)
    #:wVOVV += lib.einsum('ma,mibe->eiab', t1b,      OOVV)
    for p0, p1 in lib.prange(0, nvirb, noccb):
        wVOVV  = .5*lib.einsum('meni,mnab->eiab', eris_OVOO[:,p0:p1], taubb)
        wVOVV -= .5*lib.einsum('me,miab->eiab', Fovb[:,p0:p1], t2bb)

        tmp1bb = lib.einsum('nibf,menf->mbei', t2bb, OVOV[:,p0:p1])
        tmp1bb-= lib.einsum('nIfB,nfME->MBEI', t2ab, eris_ovOV[:,:,:,p0:p1])
        wVOVV += lib.einsum('ma,mbei->eiab', t1b, tmp1bb)
        wVOVV += lib.einsum('ma,mibe->eiab', t1b, OOVV[:,:,:,p0:p1])
        _wVOVV[p0:p1] = wVOVV
        tmp1bb = None
    eris_OVOV = eris_OVOO = eris_ovOV = None

    #:eris_OVVV = lib.unpack_tril(np.asarray(eris.OVVV).reshape(noccb*nvirb,-1)).reshape(noccb,nvirb,nvirb,nvirb)
    #:OVVV = eris_OVVV - eris_OVVV.transpose(0,3,2,1)
    #:wVOVV -= lib.einsum('MEBF,MIAF->EIAB',      OVVV, t2bb)
    #:eris_ovVV = lib.unpack_tril(np.asarray(eris.ovVV).reshape(nocca*nvira,-1)).reshape(nocca,nvira,nvirb,nvirb)
    #:wVOVV -= lib.einsum('mfBE,mIfA->EIAB', eris_ovVV, t2ab)
    #:wVOVV += eris_OVVV.transpose(2,0,3,1).conj()
    #:self.wVOVV += wVOVV - wVOVV.transpose(0,1,3,2)
    blksize = min(noccb, max(BLKMIN, int(max_memory*1e6/8/(nvirb**3*6))))
    for i0,i1 in lib.prange(0, noccb, blksize):
        wVOVV = _wVOVV[:,i0:i1]
        for p0,p1 in lib.prange(0, nocca, blksize):
            ovVV = eris.get_ovVV(slice(p0,p1))  # ovVV = eris.ovVV[p0:p1]
            wVOVV -= lib.einsum('mfBE,mIfA->EIAB', ovVV, t2ab[p0:p1,i0:i1])
            ovVV = None
        for p0,p1 in lib.prange(0, noccb, blksize):
            OVVV = eris.get_OVVV(slice(p0,p1))  # OVVV = eris.OVVV[p0:p1]
            if p0 == i0:
                wVOVV += OVVV.transpose(2,0,3,1).conj()
            OVVV = OVVV - OVVV.transpose(0,3,2,1)
            wVOVV -= lib.einsum('mebf,miaf->eiab', OVVV, t2bb[p0:p1,i0:i1])
            OVVV = None
        wVOVV = wVOVV - wVOVV.transpose(0,1,3,2)
        _wVOVV[:,i0:i1] = wVOVV

    eris_ovOV = np.asarray(eris.ovOV)
    eris_ovOO = np.asarray(eris.ovOO)
    eris_OOvv = np.asarray(eris.OOvv)
    eris_ovVO = np.asarray(eris.ovVO)
    #:self.wvOvV = lib.einsum('meNI,mNaB->eIaB', eris_ovOO, tauab)
    #:self.wvOvV -= lib.einsum('me,mIaB->eIaB', self.Fova, t2ab)
    #:tmp1ab = lib.einsum('NIBF,meNF->mBeI', t2bb, eris_ovOV)
    #:tmp1ab-= lib.einsum('nIfB,menf->mBeI', t2ab,      ovov)
    #:tmp1baab = lib.einsum('nIbF,neMF->MbeI', t2ab, eris_ovOV)
    #:tmpab = lib.einsum('ma,mBeI->eIaB', t1a, tmp1ab)
    #:tmpab+= lib.einsum('MA,MbeI->eIbA', t1b, tmp1baab)
    #:tmpab-= lib.einsum('MA,MIbe->eIbA', t1b, eris_OOvv)
    #:tmpab-= lib.einsum('ma,meBI->eIaB', t1a, eris_ovVO)
    #:self.wvOvV += tmpab
    for p0, p1 in lib.prange(0, nvira, nocca):
        wvOvV  = lib.einsum('meNI,mNaB->eIaB', eris_ovOO[:,p0:p1], tauab)
        wvOvV -= lib.einsum('me,mIaB->eIaB', Fova[:,p0:p1], t2ab)
        tmp1ab = lib.einsum('NIBF,meNF->mBeI', t2bb, eris_ovOV[:,p0:p1])
        tmp1ab-= lib.einsum('nIfB,menf->mBeI', t2ab, ovov[:,p0:p1])
        wvOvV+= lib.einsum('ma,mBeI->eIaB', t1a, tmp1ab)
        tmp1ab = None
        tmp1baab = lib.einsum('nIbF,neMF->MbeI', t2ab, eris_ovOV[:,p0:p1])
        wvOvV+= lib.einsum('MA,MbeI->eIbA', t1b, tmp1baab)
        tmp1baab = None
        wvOvV-= lib.einsum('MA,MIbe->eIbA', t1b, eris_OOvv[:,:,:,p0:p1])
        wvOvV-= lib.einsum('ma,meBI->eIaB', t1a, eris_ovVO[:,p0:p1])
        _wvOvV[p0:p1] = wvOvV
    eris_ovOV = eris_ovOO = eris_OOvv = eris_ovVO = None

    #:eris_ovvv = lib.unpack_tril(np.asarray(eris.ovvv).reshape(nocca*nvira,-1)).reshape(nocca,nvira,nvira,nvira)
    #:ovvv = eris_ovvv - eris_ovvv.transpose(0,3,2,1)
    #:self.wvOvV -= lib.einsum('mebf,mIfA->eIbA',      ovvv, t2ab)
    #:eris_ovVV = lib.unpack_tril(np.asarray(eris.ovVV).reshape(nocca*nvira,-1)).reshape(nocca,nvira,nvirb,nvirb)
    #:self.wvOvV -= lib.einsum('meBF,mIaF->eIaB', eris_ovVV, t2ab)
    #:eris_OVvv = lib.unpack_tril(np.asarray(eris.OVvv).reshape(noccb*nvirb,-1)).reshape(noccb,nvirb,nvira,nvira)
    #:self.wvOvV -= lib.einsum('MFbe,MIAF->eIbA', eris_OVvv, t2bb)
    #:self.wvOvV += eris_OVvv.transpose(2,0,3,1).conj()
    blksize = min(noccb, max(BLKMIN, int(max_memory*1e6/8/(nvira**3*6))))
    for i0,i1 in lib.prange(0, noccb, blksize):
        wvOvV = _wvOvV[:,i0:i1]
        for p0,p1 in lib.prange(0, nocca, blksize):
            ovVV = eris.get_ovVV(slice(p0,p1))  # ovVV = eris.ovVV[p0:p1]
            wvOvV -= lib.einsum('meBF,mIaF->eIaB', ovVV, t2ab[p0:p1,i0:i1])
            ovVV = None
        for p0,p1 in lib.prange(0, nocca, blksize):
            ovvv = eris.get_ovvv(slice(p0,p1))  # ovvv = eris.ovvv[p0:p1]
            ovvv = ovvv - ovvv.transpose(0,3,2,1)
            wvOvV -= lib.einsum('mebf,mIfA->eIbA',ovvv, t2ab[p0:p1,i0:i1])
            ovvv = None
        _wvOvV[:,i0:i1] = wvOvV

    blksize = min(noccb, max(BLKMIN, int(max_memory*1e6/8/(nvirb*nvira**2*3))))
    for i0,i1 in lib.prange(0, noccb, blksize):
        wvOvV = _wvOvV[:,i0:i1]
        for p0,p1 in lib.prange(0, noccb, blksize):
            OVvv = eris.get_OVvv(slice(p0,p1))  # OVvv = eris.OVvv[p0:p1]
            if p0 == i0:
                wvOvV += OVvv.transpose(2,0,3,1).conj()
            wvOvV -= lib.einsum('MFbe,MIAF->eIbA', OVvv, t2bb[p0:p1,i0:i1])
            OVvv = None
        _wvOvV[:,i0:i1] = wvOvV

    eris_ovOV = np.asarray(eris.ovOV)
    eris_OVoo = np.asarray(eris.OVoo)
    eris_ooVV = np.asarray(eris.ooVV)
    eris_OVvo = np.asarray(eris.OVvo)
    #:self.wVoVv = lib.einsum('MEni,nMbA->EiAb', eris_OVoo, tauab)
    #:self.wVoVv -= lib.einsum('ME,iMbA->EiAb', self.Fovb, t2ab)
    #:tmp1ba = lib.einsum('nibf,nfME->MbEi', t2aa, eris_ovOV)
    #:tmp1ba-= lib.einsum('iNbF,MENF->MbEi', t2ab,      OVOV)
    #:tmp1abba = lib.einsum('iNfB,mfNE->mBEi', t2ab, eris_ovOV)
    #:tmpba = lib.einsum('MA,MbEi->EiAb', t1b, tmp1ba)
    #:tmpba+= lib.einsum('ma,mBEi->EiBa', t1a, tmp1abba)
    #:tmpba-= lib.einsum('ma,miBE->EiBa', t1a, eris_ooVV)
    #:tmpba-= lib.einsum('MA,MEbi->EiAb', t1b, eris_OVvo)
    #:self.wVoVv += tmpba
    for p0, p1 in lib.prange(0, nvirb, noccb):
        wVoVv  = lib.einsum('MEni,nMbA->EiAb', eris_OVoo[:,p0:p1], tauab)
        wVoVv -= lib.einsum('ME,iMbA->EiAb', Fovb[:,p0:p1], t2ab)
        tmp1ba = lib.einsum('nibf,nfME->MbEi', t2aa, eris_ovOV[:,:,:,p0:p1])
        tmp1ba-= lib.einsum('iNbF,MENF->MbEi', t2ab, OVOV[:,p0:p1])
        wVoVv += lib.einsum('MA,MbEi->EiAb', t1b, tmp1ba)
        tmp1ba = None
        tmp1abba = lib.einsum('iNfB,mfNE->mBEi', t2ab, eris_ovOV[:,:,:,p0:p1])
        wVoVv += lib.einsum('ma,mBEi->EiBa', t1a, tmp1abba)
        tmp1abba = None
        wVoVv -= lib.einsum('ma,miBE->EiBa', t1a, eris_ooVV[:,:,:,p0:p1])
        wVoVv -= lib.einsum('MA,MEbi->EiAb', t1b, eris_OVvo[:,p0:p1])
        _wVoVv[p0:p1] = wVoVv
    eris_ovOV = eris_OVoo = eris_ooVV = eris_OVvo = None

    #:eris_OVVV = lib.unpack_tril(np.asarray(eris.OVVV).reshape(noccb*nvirb,-1)).reshape(noccb,nvirb,nvirb,nvirb)
    #:OVVV = eris_OVVV - eris_OVVV.transpose(0,3,2,1)
    #:self.wVoVv -= lib.einsum('MEBF,iMaF->EiBa',      OVVV, t2ab)
    #:eris_OVvv = lib.unpack_tril(np.asarray(eris.OVvv).reshape(noccb*nvirb,-1)).reshape(noccb,nvirb,nvira,nvira)
    #:self.wVoVv -= lib.einsum('MEbf,iMfA->EiAb', eris_OVvv, t2ab)
    #:eris_ovVV = lib.unpack_tril(np.asarray(eris.ovVV).reshape(nocca*nvira,-1)).reshape(nocca,nvira,nvirb,nvirb)
    #:self.wVoVv -= lib.einsum('mfBE,miaf->EiBa', eris_ovVV, t2aa)
    #:self.wVoVv += eris_ovVV.transpose(2,0,3,1).conj()
    blksize = min(noccb, max(BLKMIN, int(max_memory*1e6/8/(nvirb**3*6))))
    for i0,i1 in lib.prange(0, nocca, blksize):
        wVoVv = _wVoVv[:,i0:i1]
        for p0,p1 in lib.prange(0, noccb, blksize):
            OVvv = eris.get_OVvv(slice(p0,p1))  # OVvv = eris.OVvv[p0:p1]
            wVoVv -= lib.einsum('MEbf,iMfA->EiAb', OVvv, t2ab[i0:i1,p0:p1])
            OVvv = None
        for p0,p1 in lib.prange(0, noccb, blksize):
            OVVV = eris.get_OVVV(slice(p0,p1))  # OVVV = eris.OVVV[p0:p1]
            OVVV = OVVV - OVVV.transpose(0,3,2,1)
            wVoVv -= lib.einsum('MEBF,iMaF->EiBa', OVVV, t2ab[i0:i1,p0:p1])
            OVVV = None
        _wVoVv[:,i0:i1] = wVoVv

    blksize = min(nocca, max(BLKMIN, int(max_memory*1e6/8/(nvira*nvirb**2*3))))
    for i0,i1 in lib.prange(0, nocca, blksize):
        wVoVv = _wVoVv[:,i0:i1]
        for p0,p1 in lib.prange(0, nocca, blksize):
            ovVV = eris.get_ovVV(slice(p0,p1))  # ovVV = eris.ovVV[p0:p1]
            if p0 == i0:
                wVoVv += ovVV.transpose(2,0,3,1).conj()
            wVoVv -= lib.einsum('mfBE,miaf->EiBa', ovVV, t2aa[p0:p1,i0:i1])
            ovVV = None
        _wVoVv[:,i0:i1] = wVoVv
    wvvvo = _wvovv.transpose(2,0,3,1)
    wVVvo = _wVoVv.transpose(2,0,3,1)
    wvvVO = _wvOvV.transpose(2,0,3,1)
    wVVVO = _wVOVV.transpose(2,0,3,1)
    return wvvvo, wvvVO, wVVvo, wVVVO

def _get_vvVV(eris):
    if eris.vvVV is None and getattr(eris, 'VVL', None) is not None:  # DF eris
        vvL = np.asarray(eris.vvL)
        VVL = np.asarray(eris.VVL)
        vvVV = lib.dot(vvL, VVL.T)
    elif len(eris.vvVV.shape) == 2:
        vvVV = np.asarray(eris.vvVV)
    else:
        return eris.vvVV

    nvira = int(np.sqrt(vvVV.shape[0]*2))
    nvirb = int(np.sqrt(vvVV.shape[1]*2))
    vvVV1 = np.zeros((nvira**2,nvirb**2))
    vtrila = np.tril_indices(nvira)
    vtrilb = np.tril_indices(nvirb)
    lib.takebak_2d(vvVV1, vvVV, vtrila[0]*nvira+vtrila[1], vtrilb[0]*nvirb+vtrilb[1])
    lib.takebak_2d(vvVV1, vvVV, vtrila[1]*nvira+vtrila[0], vtrilb[1]*nvirb+vtrilb[0])
    lib.takebak_2d(vvVV1, vvVV, vtrila[0]*nvira+vtrila[1], vtrilb[1]*nvirb+vtrilb[0])
    lib.takebak_2d(vvVV1, vvVV, vtrila[1]*nvira+vtrila[0], vtrilb[0]*nvirb+vtrilb[1])
    return vvVV1.reshape(nvira,nvira,nvirb,nvirb)

def _get_VVVV(eris):
    if eris.VVVV is None and getattr(eris, 'VVL', None) is not None:  # DF eris
        VVL = np.asarray(eris.VVL)
        nvir = int(np.sqrt(eris.VVL.shape[0]*2))
        return ao2mo.restore(1, lib.dot(VVL, VVL.T), nvir)
    elif len(eris.VVVV.shape) == 2:
        nvir = int(np.sqrt(eris.VVVV.shape[0]*2))
        return ao2mo.restore(1, np.asarray(eris.VVVV), nvir)
    else:
        return eris.VVVV



#!/usr/bin/env python
# Copyright 2014-2022 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from pyscf import lib
from pyscf.grad import mcpdft as mcpdft_grad
from pyscf.df.grad import sacasscf as dfsacasscf_grad
from pyscf.df.grad import rhf as dfrhf_grad
from functools import partial

# I need to resolve the __init__ and get_ham_response members. Otherwise everything should be fine!
class Gradients (dfsacasscf_grad.Gradients, mcpdft_grad.Gradients):

    def __init__(self, pdft, state=None):
        self.auxbasis_response = True
        mcpdft_grad.Gradients.__init__(self, pdft, state=state)

    # TODO: rewrite the partialized fn to take the actual caller, use getattr,
    # and delete this
    def get_ham_response (self, **kwargs):
        pfn = partial (mcpdft_grad.mcpdft_HellmanFeynman_grad,
         auxbasis_response=self.auxbasis_response)
        with lib.temporary_env (mcpdft_grad, mcpdft_HellmanFeynman_grad=pfn):
            return mcpdft_grad.Gradients.get_ham_response (self, **kwargs)

    kernel = mcpdft_grad.Gradients.kernel
    get_wfn_response = mcpdft_grad.Gradients.get_wfn_response
    get_init_guess = mcpdft_grad.Gradients.get_init_guess
    project_Aop = mcpdft_grad.Gradients.project_Aop




#!/usr/bin/env python
# Copyright 2014-2021 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Tianyu Zhu <zhutianyu1991@gmail.com>
#

"""
Spin-restricted random phase approximation (direct RPA/dRPA in chemistry)
with N^4 scaling

Method:
    Main routines are based on GW-AC method described in:
    T. Zhu and G.K.-L. Chan, J. Chem. Theory. Comput. 17, 727-741 (2021)
    X. Ren et al., New J. Phys. 14, 053020 (2012)
"""

import numpy as np, scipy

from pyscf import lib
from pyscf.lib import logger
from pyscf.ao2mo import _ao2mo
from pyscf import df, scf
from pyscf.mp import dfmp2

einsum = lib.einsum

# ****************************************************************************
# core routines kernel
# ****************************************************************************

def kernel(rpa, eris=None, nw=40, x0=0.5, verbose=None):
    '''
    RPA correlation and total energy

    Args:
        cderi_ov:
            Array-like object, Cholesky decomposed ERI in OV subspace.
        nw:
            number of frequency point on imaginary axis.
        x0:
            scaling factor for frequency grid.

    Returns:
        e_tot:
            RPA total energy
        e_hf:
            EXX energy
        e_corr:
            RPA correlation energy
    '''
    log = logger.new_logger(rpa, verbose)

    if eris is None:
        eris = rpa.ao2mo()
    naux = eris.naux

    cput1 = (logger.process_clock(), logger.perf_counter())

    # Compute exact exchange energy (EXX)
    e_hf = rpa.get_e_hf()

    cput2 = cput1 = log.timer('EXX energy', *cput1)

    # Compute RPA correlation energy
    e_corr = 0
    e_ov = rpa.make_e_ov()
    f_ov = rpa.make_f_ov()
    for igrid,(omega,weigh) in enumerate(zip(*_get_scaled_legendre_roots(nw, x0))):
        diel = rpa.make_dielectric_matrix(omega, e_ov, f_ov, eris)
        factor = weigh / (2.0 * np.pi)
        e_corr += factor * np.log(np.linalg.det(np.eye(naux) - diel))
        e_corr += factor * np.trace(diel)
        diel = None

        cput2 = log.timer_debug1('RPA corr grid %d/%d'%(igrid+1,nw), *cput2)
    log.timer('RPA corr', *cput1)

    if abs(e_corr.imag) > 1e-4:
        log.warn('Non-zero imaginary part found in %s energy %s', rpa.__class__.__name__, e_corr)
    e_corr = e_corr.real

    return e_hf, e_corr

# ****************************************************************************
# frequency integral quadrature, legendre, clenshaw_curtis
# ****************************************************************************

def make_dielectric_matrix(omega, e_ov, f_ov, eris, blksize=None):
    '''
    Compute dielectric matrix at a given frequency omega

    Args:
        omega : float, frequency
        e_ov : 1D array (nocc * nvir), orbital energy differences
        eris : DF ERI object

    Returns:
        diel : 2D array (naux, naux), dielectric matrix
    '''
    assert blksize is not None

    nocc, nvir, naux = eris.nocc, eris.nvir, eris.naux

    isreal = eris.dtype == np.float64

    chi0 = (2.0 * e_ov * f_ov / (omega ** 2 + e_ov ** 2)).ravel()
    diel = np.zeros((naux, naux), dtype=eris.dtype)

    for p0,p1 in lib.prange(0, nocc*nvir, blksize):
        ovL = eris.get_ov_blk(p0,p1)
        ovL_chi = (ovL.T * chi0[p0:p1]).T
        if isreal:
            lib.ddot(ovL_chi.T, ovL, c=diel, beta=1)
        else:
            lib.dot(ovL_chi.T, ovL.conj(), c=diel, beta=1)
        ovL = ovL_chi = None

    return diel

def _get_scaled_legendre_roots(nw, x0=0.5):
    '''
    Scale nw Legendre roots, which lie in the
    interval [-1, 1], so that they lie in [0, inf)
    Ref: www.cond-mat.de/events/correl19/manuscripts/ren.pdf

    Returns:
        freqs : 1D array
        wts : 1D array
    '''
    freqs, wts = np.polynomial.legendre.leggauss(nw)
    freqs_new = x0 * (1.0 + freqs) / (1.0 - freqs)
    wts = wts * 2.0 * x0 / (1.0 - freqs)**2
    return freqs_new, wts

def _get_clenshaw_curtis_roots(nw):
    '''
    Clenshaw-Curtis quadrature on [0,inf)
    Ref: J. Chem. Phys. 132, 234114 (2010)

    Returns:
        freqs : 1D array
        wts : 1D array
    '''
    freqs = np.zeros(nw)
    wts = np.zeros(nw)
    a = 0.2
    for w in range(nw):
        t = (w + 1.0) / nw * np.pi * 0.5
        freqs[w] = a / np.tan(t)
        if w != nw - 1:
            wts[w] = a * np.pi * 0.50 / nw / (np.sin(t)**2)
        else:
            wts[w] = a * np.pi * 0.25 / nw / (np.sin(t)**2)
    return freqs[::-1], wts[::-1]


class RPA(dfmp2.DFMP2):

    _keys = {
        'mol', 'frozen',
        'with_df', 'mo_energy',
        'mo_coeff', 'mo_occ',
        'e_corr', 'e_hf', 'e_tot',
    }

    def dump_flags(self, verbose=None):
        log = logger.new_logger(self, verbose)
        log.info('')
        log.info('******** %s ********', self.__class__)
        log.info('method = %s', self.__class__.__name__)
        log.info('nocc = %d  nmo = %d', self.nocc, self.nmo)
        if self.frozen is not None:
            log.info('frozen orbitals = %s', self.frozen)
        return self

    def kernel(self, eris=None, nw=40, x0=0.5):
        '''
        The kernel function for direct RPA
        '''
        if np.iscomplexobj(self.mo_coeff):
            ''' The code runs for complex-valued orbitals but the results ain't quite right
                (very close though...). Throw an exception for now.
            '''
            raise NotImplementedError

        log = logger.new_logger(self)
        cput0 = (logger.process_clock(), logger.perf_counter())

        self.dump_flags()
        res = kernel(
            self, eris=eris, nw=nw, x0=x0, verbose=self.verbose)
        self.e_hf, self.e_corr = res

        log.timer(self.__class__.__name__, *cput0)

        self._finalize()

        return self.e_corr

    def _finalize(self):
        '''Hook for dumping results and clearing up the object.'''
        log = logger.new_logger(self)
        log.note('E(%s) = %.15g  E_corr = %.15g  E_hf = %.15g',
                 self.__class__.__name__, self.e_tot, self.e_corr, self.e_hf)

    def get_e_hf(self):
        if self.e_hf is None:
            mf = self._scf
            if isinstance(mf, scf.hf.KohnShamDFT):
                mf = mf.to_hf()
            if getattr(mf, 'with_df', None) is None:
                mf = mf.density_fit(with_df=self.with_df)
            dm = mf.make_rdm1()
            self.e_hf = mf.energy_elec(dm=dm)[0] + mf.energy_nuc()
        return self.e_hf

    def make_e_ov(self):
        '''
        Compute orbital energy differences
        '''
        log = logger.new_logger(self)
        moeocc, moevir = self.split_mo_energy()[1:3]
        e_ov = (moeocc[:,None] - moevir).ravel()

        gap = (-e_ov.max(), )
        log.info('Lowest orbital energy difference: % 6.4e', np.min(gap))

        if (np.min(gap) < 1e-3):
            log.warn('RPA code is not well-defined for degenerate systems!')
            log.warn('Lowest orbital energy difference: % 6.4e', np.min(gap))

        return e_ov

    def make_f_ov(self):
        '''
        Compute orbital occupation number differences
        '''
        focc, fvir = self.split_mo_occ()[1:3]
        return (focc[:,None] - fvir).ravel()

    def make_dielectric_matrix(self, omega, e_ov=None, f_ov=None, eris=None,
                               max_memory=None, blksize=None):
        '''
        Args:
            omega : float, frequency
            e_ov : 1D array (nocc * nvir), orbital energy differences
            mo_coeff :  (nao, nmo), mean-field mo coefficient
            cderi_ov :  (naux, nocc, nvir), Cholesky decomposed ERI in OV subspace.

        Returns:
            diel : 2D array (naux, naux), dielectric matrix
        '''
        if e_ov is None: e_ov = self.make_e_ov()
        if f_ov is None: f_ov = self.make_f_ov()
        if eris is None: eris = self.ao2mo()
        if max_memory is None: max_memory = self.max_memory

        if blksize is None:
            mem_avail = max_memory - lib.current_memory()[0]
            nocc, nvir, naux = eris.nocc, eris.nvir, eris.naux
            dsize = eris.dsize
            mem_blk = 2*naux * dsize/1e6    # ovL and ovL*chi0
            blksize = max(1, min(nocc*nvir, int(np.floor(mem_avail*0.7 / mem_blk))))
        else:
            blksize = min(blksize, e_ov.size)

        diel = make_dielectric_matrix(omega, e_ov, f_ov, eris, blksize=blksize)

        return diel


dRPA = DirectRPA = RPA


if __name__ == '__main__':
    from pyscf import gto, dft
    mol = gto.Mole()
    mol.atom = [
        [8 , (0. , 0.     , 0.)],
        [1 , (0. , -0.7571 , 0.5861)],
        [1 , (0. , 0.7571 , 0.5861)]]
    mol.basis = 'def2svp'
    mol.build()
    mol.verbose = 4

    mf = dft.RKS(mol)
    mf.xc = 'pbe'
    mf.kernel()

    rpa = RPA(mf)
    rpa.verbose = 4

    eris = rpa.ao2mo()
    e_corr_0 = rpa.kernel(eris=eris)

    assert (abs(rpa.e_corr - -0.307830040357800) < 1e-6)
    assert (abs(rpa.e_tot  - -76.26651423730257) < 1e-6)

    # Another implementation of direct RPA N^6
    e_ov = -rpa.make_e_ov()
    nov = e_ov.size
    v_ov = np.asarray(eris.get_ov_blk(0,nov).T, order='C')
    a = e_ov * np.eye(nov) + 2 * np.dot(v_ov.T, v_ov)
    b = 2 * np.dot(v_ov.T, v_ov)
    cmat = np.block([[a,b],[-b.conj(),-a.conj()]])
    ev = scipy.linalg.eig(cmat)[0]
    ev = ev.real
    ev = ev[ev>0]
    e_corr_1 = 0.5 * (np.sum(ev) - np.trace(a))

    assert abs(e_corr_0 - e_corr_1) < 1e-8
